import os
import traceback

import numpy as np
import lightkurve
from lcbuilder.helper import LcbuilderHelper
from lcbuilder.lcbuilder_class import LcBuilder
from lcbuilder.objectinfo import MissionObjectInfo
from lcbuilder.star.starinfo import StarInfo

from sherlockpipe.search import sherlock
from sherlockpipe.loading.common import load_from_yaml, get_from_dict_or_default, get_from_user_or_config_or_default, \
    get_from_dict, get_from_user_or_config, extract_custom_class
from sherlockpipe.search.sherlock_target import SherlockTarget
from os import path


"""Used to launch the SHERLOCK search and also containing few helper methods"""


def get_star_info(object_id: str, target: dict) -> StarInfo:
    """
    Reads the properties for the target star and returns a StarInfo

    :param str object_id: the target object id
    :param dict target: the dictionary containing the target definition
    :return StarInfo: returns a StarInfo
    """
    input_star_info = None
    if isinstance(target, dict) and "STAR" in target and target["STAR"] is not None:
        star_properties = target["STAR"]
        input_star_info = StarInfo(object_id=object_id, ld_coefficients=tuple(
            star_properties["LD_COEFFICIENTS"]) if "LD_COEFFICIENTS" in star_properties else None,
                                   teff=star_properties["TEFF"] if "TEFF" in star_properties else None,
                                   lum=star_properties["LUM"] if "LUM" in star_properties else None,
                                   logg=star_properties["LOGG"] if "LOGG" in star_properties else None,
                                   radius=star_properties["RADIUS"] if "RADIUS" in star_properties else None,
                                   radius_min=star_properties[
                                       "RADIUS_LOWER_ERROR"] if "RADIUS_LOWER_ERROR" in star_properties else None,
                                   radius_max=star_properties[
                                       "RADIUS_UPPER_ERROR"] if "RADIUS_UPPER_ERROR" in star_properties else None,
                                   mass=star_properties["MASS"] if "MASS" in star_properties else None,
                                   mass_min=star_properties[
                                       "MASS_LOWER_ERROR"] if "MASS_LOWER_ERROR" in star_properties else None,
                                   mass_max=star_properties[
                                       "MASS_UPPER_ERROR"] if "MASS_UPPER_ERROR" in star_properties else None,
                                   ra=star_properties["RA"] if "RA" in star_properties else None,
                                   dec=star_properties["DEC"] if "DEC" in star_properties else None)
    return input_star_info


def extract_sectors(object_info: MissionObjectInfo, cache_dir: str) -> np.ndarray:
    """
    Given the object info and the cache directory, it searches in `lightkurve` for the target pixel files and retrieves
    the sectors, quarters or campaigns either as an integer (if only one) or as a list.

    :param MissionObjectInfo object_info: the object info with the target information
    :param str cache_dir: the lightkurve cache directory
    :return object: either an integer or a list of integers
    """
    mission, mission_prefix, id_int = LcBuilder().parse_object_info(object_info.mission_id())
    lcf_search_results = lightkurve.search_targetpixelfile(object_info.mission_id(),
                                                           mission=mission, cadence="long")\
        .download_all(download_dir=cache_dir)
    sector_name, sectors = LcbuilderHelper.mission_lightkurve_sector_extraction(mission, lcf_search_results)
    return sectors


def run_search(properties: str, explore: bool, results_dir: None, cpus: int = None):
    """
    Executes the SHERLOCK search reading the given properties file.

    :param properties: the properties directory
    :param explore: whether SHERLOCK should run only the steps previous to the search.
    :param cpus: the number of processes to be used
    :param results_dir: the output directory to store the results
    :return dict: the final used properties dictionary
    """
    resources_dir = os.path.dirname(path.join(path.dirname(__file__)))
    file_dir = resources_dir + "/" + 'properties.yaml' if resources_dir != "" and resources_dir is not None \
        else 'properties.yaml'
    sherlock_user_properties = load_from_yaml(file_dir)
    user_properties = load_from_yaml(properties)
    sherlock_user_properties.update(user_properties)
    sherlock.Sherlock([], explore, sherlock_user_properties["UPDATE_OIS"],
                      sherlock_user_properties["UPDATE_FORCE"], sherlock_user_properties["UPDATE_CLEAN"],
                      results_dir=results_dir).run()
    sherlock_targets = []
    lcbuilder = LcBuilder()
    cache_dir = get_from_dict_or_default(sherlock_user_properties, "CACHE_DIR", os.path.expanduser('~') + "/")
    for target, target_configs in sherlock_user_properties["TARGETS"].items():
        try:
            aperture = get_from_dict(target_configs, "APERTURE")
            file = get_from_user_or_config(target_configs, sherlock_user_properties, "FILE")
            author = get_from_user_or_config(target_configs, sherlock_user_properties, "AUTHOR")
            star_info = get_star_info(target, target_configs)
            min_sectors = get_from_user_or_config(target_configs, sherlock_user_properties, "MIN_SECTORS")
            max_sectors = get_from_user_or_config(target_configs, sherlock_user_properties, "MAX_SECTORS")
            initial_trim = get_from_user_or_config(target_configs, sherlock_user_properties, "INITIAL_TRIM")
            initial_trim_sectors = get_from_user_or_config(target_configs, sherlock_user_properties, "INITIAL_TRIM_SECTORS")
            bin_minutes = get_from_user_or_config(target_configs, sherlock_user_properties, "BIN_MINUTES")
            mask_mode = get_from_user_or_config(target_configs, sherlock_user_properties, "MASK_MODE")
            cpu_cores = get_from_user_or_config(target_configs, sherlock_user_properties, "CPU_CORES") \
                if cpus is None else cpus
            max_runs = get_from_user_or_config(target_configs, sherlock_user_properties, "MAX_RUNS")
            period_min = get_from_user_or_config(target_configs, sherlock_user_properties, "PERIOD_MIN")
            period_max = get_from_user_or_config(target_configs, sherlock_user_properties, "PERIOD_MAX")
            period_protect = get_from_user_or_config(target_configs, sherlock_user_properties, "PERIOD_PROTECT")
            best_signal_algorithm = get_from_user_or_config(target_configs, sherlock_user_properties,
                                                            "BEST_SIGNAL_ALGORITHM")
            quorum_strength = get_from_user_or_config(target_configs, sherlock_user_properties, "QUORUM_STRENGTH")
            min_quorum = get_from_user_or_config(target_configs, sherlock_user_properties, "MIN_QUORUM")
            fit_method = get_from_user_or_config(target_configs, sherlock_user_properties, "FIT_METHOD")
            oversampling = get_from_user_or_config(target_configs, sherlock_user_properties, "OVERSAMPLING")
            t0_fit_margin = get_from_user_or_config(target_configs, sherlock_user_properties, "T0_FIT_MARGIN")
            duration_grid_step = get_from_user_or_config(target_configs, sherlock_user_properties, "DURATION_GRID_STEP")
            initial_mask = get_from_user_or_config(target_configs, sherlock_user_properties, "INITIAL_MASK")
            initial_transit_mask = get_from_user_or_config(target_configs, sherlock_user_properties,
                                                           "INITIAL_TRANSIT_MASK")
            initial_ois_mask = get_from_user_or_config(target_configs, sherlock_user_properties,
                                                           "INITIAL_OIS_MASK")
            sde_min = get_from_user_or_config(target_configs, sherlock_user_properties, "SDE_MIN")
            snr_min = get_from_user_or_config(target_configs, sherlock_user_properties, "SNR_MIN")
            custom_search_zone = extract_custom_class(
                get_from_user_or_config(target_configs, sherlock_user_properties, "CUSTOM_SEARCH_ZONE"))
            search_zone = get_from_user_or_config(target_configs, sherlock_user_properties, "SEARCH_ZONE")
            custom_transit_template = extract_custom_class(
                get_from_user_or_config(target_configs, sherlock_user_properties,
                                        "CUSTOM_TRANSIT_TEMPLATE"))
            custom_selection_algorithm = extract_custom_class(
                get_from_user_or_config(target_configs, sherlock_user_properties,
                                        "CUSTOM_SELECTION_ALGORITHM"))
            prepare_algorithm = extract_custom_class(get_from_user_or_config(target_configs, sherlock_user_properties,
                                                                             "PREPARE_ALGORITHM"))
            detrend_cores = get_from_user_or_config(target_configs, sherlock_user_properties, "DETREND_CORES") \
                if cpus is None else cpus
            detrends_number = get_from_user_or_config(target_configs, sherlock_user_properties, "DETRENDS_NUMBER")
            detrend_l_max = get_from_user_or_config(target_configs, sherlock_user_properties, "DETREND_L_MAX")
            detrend_l_min = get_from_user_or_config(target_configs, sherlock_user_properties, "DETREND_L_MIN")
            detrend_method = get_from_user_or_config(target_configs, sherlock_user_properties, "DETREND_METHOD")
            auto_detrend_period = get_from_user_or_config(target_configs, sherlock_user_properties,
                                                          "AUTO_DETREND_PERIOD")
            auto_detrend_ratio = get_from_user_or_config(target_configs, sherlock_user_properties, "AUTO_DETREND_RATIO")
            auto_detrend_method = get_from_user_or_config(target_configs, sherlock_user_properties,
                                                          "AUTO_DETREND_METHOD")
            auto_detrend_enabled = get_from_user_or_config(target_configs, sherlock_user_properties,
                                                           "AUTO_DETREND_ENABLED")
            smooth_enabled = get_from_user_or_config(target_configs, sherlock_user_properties, "INITIAL_SMOOTH_ENABLED")
            high_rms_bin_hours = get_from_user_or_config(target_configs, sherlock_user_properties,
                                                         "INITIAL_HIGH_RMS_BIN_HOURS")
            high_rms_threshold = get_from_user_or_config(target_configs, sherlock_user_properties,
                                                         "INITIAL_HIGH_RMS_THRESHOLD")
            high_rms_enabled = get_from_user_or_config(target_configs, sherlock_user_properties,
                                                       "INITIAL_HIGH_RMS_MASK")
            lower_outliers_sigma = get_from_user_or_config(target_configs, sherlock_user_properties, "LOWER_OUTLIERS_SIGMA")
            upper_outliers_sigma = get_from_user_or_config(target_configs, sherlock_user_properties, "UPPER_OUTLIERS_SIGMA")
            exptime = get_from_user_or_config(target_configs, sherlock_user_properties, "EXPTIME")
            eleanor_corr_flux = get_from_user_or_config(target_configs, sherlock_user_properties,
                                                        "ELEANOR_CORRECTED_FLUX")
            reduce_simple_oscillations = get_from_user_or_config(target_configs, sherlock_user_properties,
                                                                 "SIMPLE_OSCILLATIONS_REDUCTION")
            oscillation_snr_threshold = get_from_user_or_config(target_configs, sherlock_user_properties,
                                                                "OSCILLATIONS_MIN_SNR")
            oscillation_amplitude_threshold = get_from_user_or_config(target_configs, sherlock_user_properties,
                                                                      "OSCILLATIONS_AMPLITUDE_THRESHOLD")
            oscillation_ws_scale = get_from_user_or_config(target_configs, sherlock_user_properties,
                                                           "OSCILLATIONS_WS_PERCENT")
            oscillation_min_period = get_from_user_or_config(target_configs, sherlock_user_properties,
                                                             "OSCILLATIONS_MIN_PERIOD")
            oscillation_max_period = get_from_user_or_config(target_configs, sherlock_user_properties,
                                                             "OSCILLATIONS_MAX_PERIOD")
            exptime_binning = get_from_user_or_config(target_configs, sherlock_user_properties, "EXPTIME_BINNING")
            ignore_original = get_from_user_or_config(target_configs, sherlock_user_properties, "IGNORE_ORIGINAL")
            pickle_mode = get_from_user_or_config(target_configs, sherlock_user_properties, "PICKLE_MODE")
            mission = None
            mode = get_from_user_or_config_or_default(target_configs, sherlock_user_properties, "MODE", "GLOBAL")
            sectors = get_from_user_or_config_or_default(target_configs, sherlock_user_properties, "SECTORS", "all")
            use_harmonics_spectra = get_from_user_or_config_or_default(target_configs, sherlock_user_properties, "USE_HARMONICS_SPECTRA", False)
            truncate_border = get_from_user_or_config_or_default(target_configs, sherlock_user_properties, "TRUNCATE_BORDERS_DAYS", 0)
            min_transits_count = get_from_user_or_config_or_default(target_configs, sherlock_user_properties,
                                                                    "MIN_TRANSITS_COUNT", 1)
            compute_phase_coverage = get_from_user_or_config_or_default(target_configs, sherlock_user_properties,
                                                                        "COMPUTE_PHASE_COVERAGE", False)
            if sectors != "all" and len(np.array(sectors).shape) > 1:
                if mode == "GLOBAL" or mode == "BOTH":
                    for sectors_subset in sectors:
                        built_object_info = lcbuilder.build_object_info(target, author, sectors_subset, file, exptime,
                                                                        initial_mask, initial_transit_mask,
                                                                        star_info, aperture, eleanor_corr_flux,
                                                                        upper_outliers_sigma,
                                                                        high_rms_enabled, high_rms_threshold,
                                                                        high_rms_bin_hours,
                                                                        smooth_enabled, auto_detrend_enabled,
                                                                        auto_detrend_method,
                                                                        auto_detrend_ratio, auto_detrend_period,
                                                                        prepare_algorithm,
                                                                        reduce_simple_oscillations,
                                                                        oscillation_snr_threshold,
                                                                        oscillation_amplitude_threshold,
                                                                        oscillation_ws_scale,
                                                                        oscillation_min_period, oscillation_max_period,
                                                                        exptime_binning, truncate_border,
                                                                        lower_outliers_sigma=lower_outliers_sigma,
                                                                        initial_trim=initial_trim,
                                                                        initial_trim_sectors=initial_trim_sectors)
                        sherlock_target = SherlockTarget(built_object_info,
                                                         detrend_method, detrend_l_min, detrend_l_max, detrends_number,
                                                         detrend_cores,
                                                         custom_selection_algorithm,
                                                         custom_transit_template,
                                                         search_zone, custom_search_zone,
                                                         snr_min, sde_min,
                                                         min_sectors, max_sectors,
                                                         bin_minutes,
                                                         mask_mode,
                                                         cpu_cores, max_runs, period_min,
                                                         period_max, period_protect, best_signal_algorithm,
                                                         quorum_strength,
                                                         min_quorum, fit_method, oversampling,
                                                         t0_fit_margin, duration_grid_step, properties, cache_dir,
                                                         ignore_original,
                                                         pickle_mode, use_harmonics_spectra, ois_mask=initial_ois_mask,
                                                         min_transits_count=min_transits_count,
                                                         compute_phase_coverage=compute_phase_coverage)
                        sherlock_targets.append(sherlock_target)
                if mode == "SECTOR" or mode == "BOTH" and isinstance(built_object_info, MissionObjectInfo.MissionObjectInfo):
                    if sectors == 'all':
                        sectors = extract_sectors(built_object_info, cache_dir)
                    sectors_unique = np.unique(np.array(sectors).flatten())
                    for sector in sectors_unique:
                        object_info = lcbuilder.build_object_info(target, author, [sector], file, exptime,
                                                                  initial_mask, initial_transit_mask,
                                                                  star_info, aperture, eleanor_corr_flux,
                                                                  upper_outliers_sigma,
                                                                  high_rms_enabled, high_rms_threshold,
                                                                  high_rms_bin_hours,
                                                                  smooth_enabled, auto_detrend_enabled,
                                                                  auto_detrend_method,
                                                                  auto_detrend_ratio, auto_detrend_period,
                                                                  prepare_algorithm,
                                                                  reduce_simple_oscillations, oscillation_snr_threshold,
                                                                  oscillation_amplitude_threshold, oscillation_ws_scale,
                                                                  oscillation_min_period, oscillation_max_period,
                                                                  exptime_binning, truncate_border,
                                                                  lower_outliers_sigma=lower_outliers_sigma,
                                                                  initial_trim=initial_trim,
                                                                  initial_trim_sectors=initial_trim_sectors)
                        sherlock_target = SherlockTarget(object_info,
                                                         detrend_method, detrend_l_min, detrend_l_max, detrends_number,
                                                         detrend_cores,
                                                         custom_selection_algorithm,
                                                         custom_transit_template,
                                                         search_zone, custom_search_zone,
                                                         snr_min, sde_min,
                                                         min_sectors, max_sectors,
                                                         bin_minutes,
                                                         mask_mode,
                                                         cpu_cores, max_runs, period_min,
                                                         period_max, period_protect, best_signal_algorithm,
                                                         quorum_strength,
                                                         min_quorum, fit_method, oversampling,
                                                         t0_fit_margin, duration_grid_step, properties, cache_dir,
                                                         ignore_original,
                                                         pickle_mode, use_harmonics_spectra, ois_mask=initial_ois_mask,
                                                         min_transits_count=min_transits_count,
                                                         compute_phase_coverage=compute_phase_coverage)
                        sherlock_targets.append(sherlock_target)
            else:
                built_object_info = lcbuilder.build_object_info(target, author, sectors, file, exptime,
                                                                initial_mask, initial_transit_mask,
                                                                star_info, aperture, eleanor_corr_flux,
                                                                upper_outliers_sigma,
                                                                high_rms_enabled, high_rms_threshold,
                                                                high_rms_bin_hours,
                                                                smooth_enabled, auto_detrend_enabled,
                                                                auto_detrend_method,
                                                                auto_detrend_ratio, auto_detrend_period,
                                                                prepare_algorithm,
                                                                reduce_simple_oscillations, oscillation_snr_threshold,
                                                                oscillation_amplitude_threshold, oscillation_ws_scale,
                                                                oscillation_min_period, oscillation_max_period,
                                                                exptime_binning, truncate_border,
                                                                lower_outliers_sigma=lower_outliers_sigma,
                                                                initial_trim=initial_trim,
                                                                initial_trim_sectors=initial_trim_sectors)
                sherlock_target = SherlockTarget(built_object_info,
                                                 detrend_method, detrend_l_min, detrend_l_max, detrends_number,
                                                 detrend_cores,
                                                 custom_selection_algorithm,
                                                 custom_transit_template,
                                                 search_zone, custom_search_zone,
                                                 snr_min, sde_min,
                                                 min_sectors, max_sectors,
                                                 bin_minutes,
                                                 mask_mode,
                                                 cpu_cores, max_runs, period_min,
                                                 period_max, period_protect, best_signal_algorithm, quorum_strength,
                                                 min_quorum, fit_method, oversampling,
                                                 t0_fit_margin, duration_grid_step, properties, cache_dir,
                                                 ignore_original, pickle_mode, use_harmonics_spectra,
                                                 ois_mask=initial_ois_mask, min_transits_count=min_transits_count,
                                                 compute_phase_coverage=compute_phase_coverage)
                if mode == "GLOBAL" or mode == "BOTH":
                    sherlock_targets.append(sherlock_target)
                if mode == "SECTOR" or mode == "BOTH" and isinstance(built_object_info, MissionObjectInfo.MissionObjectInfo):
                    if sectors == 'all':
                        sectors = extract_sectors(built_object_info, cache_dir)
                    for sector in sectors:
                        object_info = lcbuilder.build_object_info(target, author, [sector], file, exptime,
                                                                  initial_mask, initial_transit_mask,
                                                                  star_info, aperture, eleanor_corr_flux,
                                                                  upper_outliers_sigma,
                                                                  high_rms_enabled, high_rms_threshold,
                                                                  high_rms_bin_hours,
                                                                  smooth_enabled, auto_detrend_enabled,
                                                                  auto_detrend_method,
                                                                  auto_detrend_ratio, auto_detrend_period,
                                                                  prepare_algorithm,
                                                                  reduce_simple_oscillations, oscillation_snr_threshold,
                                                                  oscillation_amplitude_threshold, oscillation_ws_scale,
                                                                  oscillation_min_period, oscillation_max_period,
                                                                  exptime_binning, truncate_border,
                                                                  lower_outliers_sigma=lower_outliers_sigma,
                                                                  initial_trim=initial_trim,
                                                                  initial_trim_sectors=initial_trim_sectors)
                        sherlock_target = SherlockTarget(object_info,
                                                         detrend_method, detrend_l_min, detrend_l_max, detrends_number,
                                                         detrend_cores,
                                                         custom_selection_algorithm,
                                                         custom_transit_template,
                                                         search_zone, custom_search_zone,
                                                         snr_min, sde_min,
                                                         min_sectors, max_sectors,
                                                         bin_minutes,
                                                         mask_mode,
                                                         cpu_cores, max_runs, period_min,
                                                         period_max, period_protect, best_signal_algorithm,
                                                         quorum_strength,
                                                         min_quorum, fit_method, oversampling,
                                                         t0_fit_margin, duration_grid_step, properties, cache_dir,
                                                         ignore_original,
                                                         pickle_mode, use_harmonics_spectra, ois_mask=initial_ois_mask,
                                                         min_transits_count=min_transits_count,
                                                         compute_phase_coverage=compute_phase_coverage)
                        sherlock_targets.append(sherlock_target)
                if mode != "GLOBAL" and mode != "BOTH" and not (mode == "SECTOR" or mode == "BOTH" and
                                                                isinstance(built_object_info, MissionObjectInfo.MissionObjectInfo)):
                    raise ValueError("Not a valid run mode: " + str(mode) + " for target: " + str(target))
        except Exception as e:
            print("Error found for target " + target)
            traceback.print_exc()
            print("Continuing with the target list")
    sherlock.Sherlock(sherlock_targets, explore, cache_dir=cache_dir, results_dir=results_dir).run()
    return sherlock_user_properties



import itertools
import logging
import math
import os
import sys
import time
import warnings
from multiprocessing import Pool

from numba import njit
from pytransit import QuadraticModel
import batman
import ellc
import numpy as np
import astropy.constants as ac
import astropy.units as u
import wotan
from lcbuilder.lcbuilder_class import LcBuilder
from scipy import stats
from scipy.interpolate import interp1d
from scipy.signal import argrelextrema
import matplotlib.pyplot as plt

G = 6.674e-11  # m3 kg-1 s-2
AU_TO_RSUN = 215.032
Msolar_to_kg = 2.e30
Mearth_to_kg = 5.972e24
M_earth_to_M_sun = Mearth_to_kg / Msolar_to_kg
R_earth_to_R_sun = 0.009175


class ExoMoonLeastSquares:
    def __init__(self, object_dir, cpus, star_mass, star_radius, ab, planet_radius, planet_period, planet_t0, planet_duration, planet_semimajor_axis, planet_inc, planet_ecc,
             planet_arg_periastron, planet_impact_param, min_depth_rms, max_radius, t0s, time, flux,
             period_grid_size=2000, radius_grid_size=10):
        self.object_dir = object_dir
        self.cpus = cpus
        self.star_mass = star_mass
        self.star_radius = star_radius
        self.ab = ab
        self.planet_radius = planet_radius
        self.planet_period = planet_period
        self.planet_t0 = planet_t0
        self.planet_duration = planet_duration
        self.planet_semimajor_axis = planet_semimajor_axis
        self.planet_inc = planet_inc
        self.planet_ecc = planet_ecc
        self.planet_arg_periastron = planet_arg_periastron
        self.planet_impact_param = planet_impact_param
        self.time = time
        self.flux = flux
        self.t0s = t0s
        self.min_depth_rms = min_depth_rms
        self.max_radius = max_radius
        self.period_grid_size = period_grid_size
        self.radius_grid_size = radius_grid_size

    @staticmethod
    def compute_semimajor_axis(major_mass, minor_period):
        period_seconds = minor_period * 24. * 3600.
        mass_kg = major_mass * Msolar_to_kg
        a1 = (G * mass_kg * period_seconds ** 2 / 4. / (np.pi ** 2)) ** (1. / 3.)
        return a1 / 1.496e11

    @staticmethod
    def compute_hill_radius(major_mass, minor_mass, semimajor_axis, eccentricity=0):
        """
        :param major_mass: The main body mass
        :param minor_mass: The minor body mass
        :param semimajor_axis: The minor body semimajor axis in AU.
        :param eccentricity: the planet eccentricity
        :return: the hill radius of the minor body in the same units than the semimajor_axis
        """
        return AU_TO_RSUN * semimajor_axis * (1 - eccentricity) * (minor_mass / (3 * major_mass) ** (1 / 3))

    @staticmethod
    def au_to_period(mass, au):
        """
        Calculates the orbital period for the semi-major axis assuming a circular orbit.
        :param mass: the stellar mass
        :param au: the semi-major axis in astronomical units.
        :return: the period in days
        """
        mass_kg = mass * 2.e30
        a = au * 1.496e11
        return ((a ** 3) * 4 * (np.pi ** 2) / G / mass_kg) ** (1. / 2.) / 3600 / 24

    @staticmethod
    def compute_transit_duration(star_radius,
                                 transiting_body_semimajor_axis, transit_period, transiting_body_radius,
                                 impact_parameter=0):
        """

        :param star_radius: star radius
        :param transiting_body_semimajor_axis: orbit semimajor axis
        :param transit_period: in days
        :param transiting_body_radius: transiting body radius
        :param impact_parameter:
        :return:
        """
        return transit_period / np.pi * np.arcsin(np.sqrt((star_radius + transiting_body_radius) ** 2 - (impact_parameter * star_radius) ** 2) / transiting_body_semimajor_axis)
        #return 2 * moon_semimajor_axis / (planet_semimajor_axis * 2 * np.pi) * planet_period

    @staticmethod
    def compute_moon_period_grid(min, max, mode="lin", samples=10000):
        if "log" == mode:
            return np.logspace(math.log(min, 10), math.log(max, 10), samples, base=10)
        else:
            return np.linspace(min, max, samples)

    def subtract_planet_transit(self, ab, star_radius, star_mass, time, flux, planet_radius, planet_t0,
                                planet_period, planet_inc=90):
        P1 = planet_period * u.day
        a = np.cbrt((ac.G * star_mass * u.M_sun * P1 ** 2) / (4 * np.pi ** 2)).to(u.au)
        model = ellc.lc(
            t_obs=time,
            radius_1=(star_radius * u.R_sun).to(u.au) / a,  # star radius convert from AU to in units of a
            radius_2=(planet_radius * u.R_earth).to(u.au) / a,
            # convert from Rearth (equatorial) into AU and then into units of a
            sbratio=0,
            incl=planet_inc,
            light_3=0,
            t_zero=planet_t0,
            period=planet_period,
            a=None,
            q=1e-6,
            f_c=None, f_s=None,
            ldc_1=ab, ldc_2=None,
            gdc_1=None, gdc_2=None,
            didt=None,
            domdt=None,
            rotfac_1=1, rotfac_2=1,
            hf_1=1.5, hf_2=1.5,
            bfac_1=None, bfac_2=None,
            heat_1=None, heat_2=None,
            lambda_1=None, lambda_2=None,
            vsini_1=None, vsini_2=None,
            t_exp=None, n_int=None,
            grid_1='default', grid_2='default',
            ld_1='quad', ld_2=None,
            shape_1='sphere', shape_2='sphere',
            spots_1=None, spots_2=None,
            exact_grav=False, verbose=1)
        return flux - model + 1

    @staticmethod
    #@njit(fastmath=True, parallel=False)
    def compute_moon_transit_scenarios(time, flux, planet_t0, moon_initial_alpha, moon_period, moon_orbit_ranges,
                                       moon_orbit_transit_length, moon_transit_duration):
        #TODO need to take into account "prograde" or "retrograde" orbit
        orbit_scenarios = None
        for moon_orbit_range in moon_orbit_ranges:
            t0 = moon_orbit_range[0]
            t1 = moon_orbit_range[1]
            phase_delta = (t0 - planet_t0) % moon_period * 2 * np.pi
            alpha = (moon_initial_alpha + phase_delta) % (2 * np.pi)
            time_alpha = np.cos(alpha) * moon_orbit_transit_length / 2
            moon_t1 = t1 + time_alpha
            time_args = np.argwhere((time > moon_t1) & (time < moon_t1 + moon_transit_duration)).flatten()
            #TODO we'd need to fill measurement gaps (detected from the time array)
            time_moon_transit = time[time_args]
            flux_moon_transit = flux[time_args]
            time_moon_transit = time_moon_transit - (moon_t1 + moon_transit_duration / 2)
            # fig_transit, axs = plt.subplots(1, 1, figsize=(8, 8))
            # axs.plot(time_moon_transit, flux_moon_transit, color='gray', alpha=1, rasterized=True,
            #          label="Flux Transit ")
            # axs.set_title("Residuals")
            # axs.set_xlabel('Time')
            # axs.set_ylabel('Flux')
            # fig_transit.show()
            if len(time_moon_transit) > 0:
                if orbit_scenarios is None:
                    orbit_scenarios = [[alpha, time_moon_transit, flux_moon_transit]]
                orbit_scenarios.append([alpha, time_moon_transit, flux_moon_transit])
        return orbit_scenarios

    def search(self, search_input, return_lc=False):
        logging.info("Searching for period=%.5fd and alpha=%.2frad", search_input.moon_period, search_input.moon_alpha)
        planet_duration = self.compute_transit_duration(self.star_radius, self.planet_semimajor_axis * AU_TO_RSUN,
                                                        self.planet_period, self.planet_radius * R_earth_to_R_sun,
                                                        search_input.impact_param)
        moon_semimajor_axis = self.compute_semimajor_axis(planet_mass * M_earth_to_M_sun, search_input.moon_period)
        moon_orbit_transit_duration = self.compute_transit_duration(self.star_radius,
                                                                  self.planet_semimajor_axis * AU_TO_RSUN,
                                                                  self.planet_period,
                                                                  moon_semimajor_axis * AU_TO_RSUN,
                                                                    search_input.impact_param)
        moon_transit_length = self.compute_transit_duration(self.star_radius, self.planet_semimajor_axis * AU_TO_RSUN,
                                                            self.planet_period, 1 * R_earth_to_R_sun)
        # TODO we probably need to define left_transit_length and right_transit_length depending on moon orbit parameters
        moon_orbit_tokens = [[t0, t0 - planet_duration / 2] for t0 in self.t0s]
        transit_scenarios = ExoMoonLeastSquares.compute_moon_transit_scenarios(self.time, self.flux, self.planet_t0, search_input.moon_alpha,
                                                                search_input.moon_period, moon_orbit_tokens,
                                                                moon_orbit_transit_duration, moon_transit_length)
        scenario_time = []
        scenario_flux = []
        for normalized_moon_transit_scenario in transit_scenarios:
            scenario_time = np.concatenate((scenario_time, normalized_moon_transit_scenario[1].flatten()))
            scenario_flux = np.concatenate((scenario_flux, normalized_moon_transit_scenario[2].flatten()))
        sorted_time_args = np.argsort(scenario_time)
        scenario_time = scenario_time[sorted_time_args]
        scenario_flux = scenario_flux[sorted_time_args]
        outliers_args = ExoMoonLeastSquares.remove_outliers(scenario_flux, sigma_lower=float('inf'), sigma_upper=3)
        scenario_time = scenario_time[~outliers_args].flatten()
        scenario_flux = scenario_flux[~outliers_args].flatten()
        interpolated = interp1d(np.arange(len(self.model)), self.model, axis=0, fill_value='extrapolate')
        model_sample = interpolated(np.linspace(0, len(self.model), len(scenario_time)))
        # fig_transit, axs = plt.subplots(1, 1, figsize=(8, 8))
        # axs.scatter(scenario_time, scenario_flux, color='gray', alpha=0.4, rasterized=True, label="Flux Transit ")
        # axs.plot(scenario_time, model_sample, color='red', alpha=1, rasterized=True, label="Flux Transit ")
        # axs.set_title("Residuals")
        # axs.set_xlabel('Time')
        # axs.set_ylabel('Flux')
        # fig_transit.show()
        residual_calculation, residual_baseline, residual_radius, residual_model = self.calculate_residuals(scenario_time, scenario_flux,
                                                                                         model_sample, self.min_radius,
                                                                                         self.max_radius,
                                                                                         self.radius_grid_size)
        if return_lc:
            return residual_calculation, residual_baseline, residual_radius, scenario_time, scenario_flux, residual_model
        else:
            return residual_calculation, residual_baseline, residual_radius

    @staticmethod
    def spectra(chi2, oversampling_factor=1, kernel_size=30):
        SR = np.min(chi2) / chi2
        SDE_raw = (1 - np.mean(SR)) / np.std(SR)

        # Scale SDE_power from 0 to SDE_raw
        power_raw = SR - np.mean(SR)  # shift down to the mean being zero
        scale = SDE_raw / np.max(power_raw)  # scale factor to touch max=SDE_raw
        power_raw = power_raw * scale

        # Detrended SDE, named "power"
        kernel = oversampling_factor * kernel_size
        if kernel % 2 == 0:
            kernel = kernel + 1
        if len(power_raw) > 2 * kernel:
            my_median = ExoMoonLeastSquares.running_median(power_raw, kernel)
            power = power_raw - my_median
            # Re-normalize to range between median = 0 and peak = SDE
            # shift down to the mean being zero
            power = power - np.mean(power)
            SDE = np.max(power / np.std(power))
            # scale factor to touch max=SDE
            scale = SDE / np.max(power)
            power = power * scale
        else:
            power = power_raw
            SDE = SDE_raw

        return SR, power_raw, power, SDE_raw, SDE

    @staticmethod
    @njit(fastmath=True, parallel=False)
    def calculate_residuals(time, flux, model_sample, min_radius, max_radius, radius_grid_size):
        # TODO adjusting model to minimum flux value this might get improved by several scalations of min_flux
        best_residual = np.inf
        best_radius = min_radius
        best_model = model_sample
        model_baseline = np.full(len(model_sample), 1)
        residuals_baseline = np.sum((flux - model_baseline) ** 2) ** 0.5
        #radius_from_mean = np.sqrt((1 - np.mean(flux)) * (1.3**2)) / 0.00975
        for radius in np.linspace(min_radius, max_radius, radius_grid_size):
            depth = ((radius * R_earth_to_R_sun) ** 2) / star_radius ** 2
            flux_at_middle = 1 - depth
            model_sample_scaled = np.copy(model_sample)
            model_sample_scaled[model_sample_scaled < 1] = model_sample_scaled[model_sample_scaled < 1] * (
                        flux_at_middle / np.min(model_sample))
            radius_residuals = np.sum((flux - model_sample_scaled) ** 2)
            if radius_residuals < best_residual:
                best_residual = radius_residuals
                best_radius = radius
                best_model = model_sample_scaled
        # fig_transit, axs = plt.subplots(1, 1, figsize=(8, 8))
        # axs.scatter(time, flux, color='gray', alpha=0.4, rasterized=True, label="Flux Transit ")
        # axs.plot(time, best_model, color='red', alpha=1, rasterized=True, label="Flux Transit ")
        # axs.set_title("Residuals")
        # axs.set_xlabel('Time')
        # axs.set_ylabel('Flux')
        # bin_means, bin_edges, binnumber = stats.binned_statistic(time,
        #                                                          flux,
        #                                                          statistic='mean', bins=25)
        # bin_stds, _, _ = stats.binned_statistic(time,
        #                                         flux, statistic='std', bins=25)
        # bin_width = (bin_edges[1] - bin_edges[0])
        # bin_centers = bin_edges[1:] - bin_width / 2
        # axs.errorbar(bin_centers, bin_means, yerr=bin_stds / 2, xerr=bin_width / 2, marker='o', markersize=4,
        #              color='darkorange', alpha=1, linestyle='none')
        # fig_transit.show()
        return best_residual, residuals_baseline, best_radius, best_model

    @staticmethod
    def running_median(data, kernel):
        """Returns sliding median of width 'kernel' and same length as data """
        idx = np.arange(kernel) + np.arange(len(data) - kernel + 1)[:, None]
        med = np.median(data[idx], axis=1)

        # Append the first/last value at the beginning/end to match the length of
        # data and returned median
        first_values = med[0]
        last_values = med[-1]
        missing_values = len(data) - len(med)
        values_front = int(missing_values * 0.5)
        values_end = missing_values - values_front
        med = np.append(np.full(values_front, first_values), med)
        med = np.append(med, np.full(values_end, last_values))
        return med

    @staticmethod
    def remove_outliers(
        flux, sigma=5.0, sigma_lower=None, sigma_upper=None, **kwargs
    ):
        # The import time for `sigma_clip` is somehow very slow, so we use
        # a local import here.
        from astropy.stats.sigma_clipping import sigma_clip

        # First, we create the outlier mask using AstroPy's sigma_clip function
        with warnings.catch_warnings():  # Ignore warnings due to NaNs or Infs
            warnings.simplefilter("ignore")
            outlier_mask = sigma_clip(
                data=flux,
                sigma=sigma,
                sigma_lower=sigma_lower,
                sigma_upper=sigma_upper,
                **kwargs,
            ).mask
        # Second, we return the masked light curve and optionally the mask itself
        return outlier_mask

    def inject_moon(self, time, flux, t0s, planet_mass, planet_semimajor_axis, planet_ecc, moon_radius, moon_period, initial_alpha=0):
        logging.info("Injecting moon with radius of  %.2fR_e, %.2fdays and %.2frad", moon_radius, moon_period, initial_alpha)
        moon_semimajor_axis = self.compute_semimajor_axis(planet_mass * M_earth_to_M_sun, moon_period)
        moon_transit_duration = self.compute_transit_duration(self.star_radius,
                                                              planet_semimajor_axis * AU_TO_RSUN,
                                                              self.planet_period,
                                                              moon_radius * R_earth_to_R_sun)
        moon_orbit_transit_duration = self.compute_transit_duration(self.star_radius,
                                                                    planet_semimajor_axis * AU_TO_RSUN,
                                                                    self.planet_period,
                                                                    moon_semimajor_axis * AU_TO_RSUN)
        first_t0 = t0s[0]
        subtracted_planet_flux = self.subtract_planet_transit(self.ab, self.star_radius, self.star_mass, time,
                                                              flux, self.planet_radius, self.planet_t0,
                                                              self.planet_period, self.planet_inc)
        for t0 in t0s:
            phase_delta = ((t0 - first_t0) % moon_period) * 2 * np.pi
            moon_phase = (initial_alpha + phase_delta) % (2 * np.pi)
            moon_tau = np.cos(moon_phase)
            moon_t0 = t0 + moon_tau * moon_orbit_transit_duration / 2
            time_transit = time[(moon_t0 - moon_transit_duration / 2 < time) & (time < moon_t0 + moon_transit_duration / 2)]
            if len(time_transit) == 0:
                continue
            ma = batman.TransitParams()
            ma.t0 = moon_t0  # time of inferior conjunction
            ma.per = self.planet_period  # orbital period, use Earth as a reference
            ma.rp = moon_radius * R_earth_to_R_sun / self.star_radius  # planet radius (in units of stellar radii)
            ma.a = planet_semimajor_axis * AU_TO_RSUN / self.star_radius  # semi-major axis (in units of stellar radii)
            ma.inc = 90  # orbital inclination (in degrees)
            ma.ecc = planet_ecc  # eccentricity
            ma.w = 0  # longitude of periastron (in degrees)
            ma.u = self.ab  # limb darkening coefficients
            ma.limb_dark = "quadratic"  # limb darkening model
            m = batman.TransitModel(ma, time_transit)  # initializes model
            model = m.light_curve(ma)  # calculates light curve
            fig_transit, axs = plt.subplots(3, 1, figsize=(16, 10))
            axs[0].plot(time_transit, flux[
                (moon_t0 - moon_transit_duration / 2 < time) & (time < moon_t0 + moon_transit_duration / 2)],
                        color='gray', alpha=1, rasterized=True, label="Flux")
            axs[0].set_title(
                "Injected transit in t0 " + str(t0) + " with moon t0=" + str(moon_t0) + " and phase " + str(moon_phase))
            axs[0].set_xlabel('Time')
            axs[0].set_ylabel('Flux')
            axs[0].plot(time_transit, model, color='red', alpha=1, rasterized=True, label="Model")
            flux[(moon_t0 - moon_transit_duration / 2 < time) & (time < moon_t0 + moon_transit_duration / 2)] = \
                flux[(moon_t0 - moon_transit_duration / 2 < time) & (time < moon_t0 + moon_transit_duration / 2)] + model - 1
            subtracted_planet_flux[(moon_t0 - moon_transit_duration / 2 < time) & (time < moon_t0 + moon_transit_duration / 2)] = \
                subtracted_planet_flux[(moon_t0 - moon_transit_duration / 2 < time) & (time < moon_t0 + moon_transit_duration / 2)] + model - 1
            axs[1].plot(time_transit, flux[
                (moon_t0 - moon_transit_duration / 2 < time) & (time < moon_t0 + moon_transit_duration / 2)],
                        color='gray', alpha=1, rasterized=True, label="Flux")
            axs[1].set_title(
                "Injected transit in t0 " + str(t0) + " with moon t0=" + str(moon_t0) + " and phase " + str(moon_phase))
            axs[1].set_xlabel('Time')
            axs[1].set_ylabel('Flux')
            axs[1].plot(time_transit, model, color='red', alpha=1, rasterized=True, label="Model")
            axs[2].plot(time_transit, subtracted_planet_flux[
                (moon_t0 - moon_transit_duration / 2 < time) & (time < moon_t0 + moon_transit_duration / 2)])
            axs[2].set_title(
                "Injected transit in t0 " + str(t0) + " with moon t0=" + str(moon_t0) + " and phase " + str(moon_phase))
            axs[2].set_xlabel('Time')
            axs[2].set_ylabel('Flux')
            axs[2].plot(time_transit, model, color='red', alpha=1, rasterized=True, label="Model")
            fig_transit.show()
        return flux

    def remove_non_transit_flux(self, time, flux, t0s, max_planet_mass):
        max_period = self.au_to_period(max_planet_mass * M_earth_to_M_sun, self.compute_hill_radius(self.star_mass, max_planet_mass * M_earth_to_M_sun, self.planet_semimajor_axis))
        moon_semimajor_axis = self.compute_semimajor_axis(planet_mass * M_earth_to_M_sun, max_period)
        moon_orbit_transit_duration = self.compute_transit_duration(self.star_radius,
                                                              self.planet_semimajor_axis * AU_TO_RSUN,
                                                              self.planet_period,
                                                              moon_semimajor_axis * AU_TO_RSUN)
        flux_mask = []
        for t0 in t0s:
            flux_mask = np.concatenate((flux_mask, np.argwhere(
                (time > t0 - moon_orbit_transit_duration) & (time < t0 + moon_orbit_transit_duration)).flatten()))
        flux_mask = flux_mask.astype(int)
        time = time[flux_mask]
        flux = flux[flux_mask]
        flux_measurements = []
        for t0 in t0s:
            flux_measurements.append(len(flux[(time > t0 - moon_orbit_transit_duration) & (time < t0 + moon_orbit_transit_duration)]))
        typical_flux_measurement_length = np.median(flux_measurements)
        i = 0
        for flux_measurement in flux_measurements:
            if flux_measurement < typical_flux_measurement_length:
                flux_mask = np.argwhere((time <= t0s[i] - moon_orbit_transit_duration) | (time >= t0s[i] + moon_orbit_transit_duration)).flatten()
                time = time[flux_mask]
                flux = flux[flux_mask]
            i = i + 1
        return time, flux

    def run(self):
        planet_mass_grid = self.planet_mass_grid
        moon_inc_grid = self.moon_inc_grid
        moon_ecc_grid = self.moon_ecc_grid
        moon_arg_periastron_grid = self.moon_arg_periastron_grid
        self.time, self.flux = self.remove_non_transit_flux(self.time, self.flux, self.t0s, np.max(planet_mass_grid))
        rms = np.std(self.flux)
        depth = rms * self.min_depth_rms
        self.min_radius = np.sqrt(depth * ((self.star_radius / R_earth_to_R_sun) ** 2))
        self.flux = self.subtract_planet_transit(self.ab, self.star_radius, self.star_mass, self.time, self.flux,
                                            self.planet_radius, self.planet_t0, self.planet_period, self.planet_inc)
        time_model = np.arange(0.4, 0.6, 0.0001)
        ma = batman.TransitParams()
        ma.t0 = 0.5  # time of inferior conjunction
        ma.per = 1  # orbital period, use Earth as a reference
        ma.rp = self.min_radius * R_earth_to_R_sun / self.star_radius  # planet radius (in units of stellar radii)
        ma.a = self.planet_semimajor_axis * AU_TO_RSUN / self.star_radius  # semi-major axis (in units of stellar radii)
        ma.inc = 90  # orbital inclination (in degrees)
        ma.ecc = planet_ecc  # eccentricity
        ma.w = 0  # longitude of periastron (in degrees)
        ma.u = self.ab  # limb darkening coefficients
        ma.limb_dark = "quadratic"  # limb darkening model
        m = batman.TransitModel(ma, time_model)  # initializes model
        self.model = m.light_curve(ma)
        self.model = self.model[self.model < 1]
        # baseline_model = np.full(len(model), 1)
        # model = np.append(baseline_model, model)
        # model = np.append(model, baseline_model)
        search_inputs = []
        alpha_grid_size = 25
        alpha_grid = np.linspace(0, np.pi * 2 - np.pi * 2 / 25, alpha_grid_size)
        #alpha_grid = [0.0]
        impact_params_grid = [0]
        for planet_mass in planet_mass_grid:
            min_period = 0.5 # TODO compute this value somehow
            max_period = self.au_to_period(planet_mass * M_earth_to_M_sun, self.compute_hill_radius(self.star_mass, planet_mass * M_earth_to_M_sun, self.planet_semimajor_axis))
            period_grid = self.compute_moon_period_grid(min_period, max_period, samples=self.period_grid_size, mode="log")
            #period_grid = [2.0]
            for moon_inc in moon_inc_grid:
                for moon_ecc in moon_ecc_grid:
                    for moon_arg_periastron in moon_arg_periastron_grid:
                        for impact_param in impact_params_grid:
                            for moon_period in period_grid:
                                for moon_initial_alpha in alpha_grid:
                                    #TODO moon_orbit_ranges should use moon_radius ?
                                    search_inputs.append(SearchInput(
                                        moon_period, moon_initial_alpha, moon_ecc, moon_inc, moon_arg_periastron,
                                        self.min_radius, impact_param))
        with Pool(processes=self.cpus) as pool:
            all_residuals = pool.map(self.search, search_inputs)
        best_residuals_per_scenarios = []
        best_residuals_spectra = []
        best_residuals_spectra_values = []
        best_residual_per_period = np.inf
        best_residual_per_period_values = []
        residuals_matrix = np.full((alpha_grid_size, len(period_grid)), np.nan)
        for i in np.arange(0, len(search_inputs)):
            if i % 25 == 0 and i != 0:
                best_residuals_spectra = best_residuals_spectra + [best_residual_per_period]
                best_residuals_spectra_values = best_residuals_spectra_values + best_residual_per_period_values
                best_residual_per_period = np.inf
                best_residual_per_period_values = []
            all_residual = all_residuals[i]
            residuals = all_residual[0]
            residuals_baseline = all_residual[1]
            radius = all_residual[2]
            moon_period = search_inputs[i].moon_period
            moon_initial_alpha = search_inputs[i].moon_alpha
            best_residuals_per_scenarios.append(
                [moon_period, moon_initial_alpha, residuals * self.min_radius / radius, radius])
            residuals_matrix[i % 25][
                i // 25] = residuals * self.min_radius / radius if residuals < residuals_baseline else np.nan
            if best_residual_per_period > residuals and residuals < residuals_baseline:
                best_residual_per_period = residuals * self.min_radius / radius
                best_residual_per_period_values = [residuals_baseline, radius, moon_period, moon_initial_alpha]
        best_residuals_spectra = best_residuals_spectra + [best_residual_per_period]
        best_residuals_spectra = np.array(best_residuals_spectra)
        best_residuals_spectra[np.isnan(best_residuals_spectra)] = np.nanmax(best_residuals_spectra)
        best_residuals_spectra_values = best_residuals_spectra_values + best_residual_per_period_values
        residuals_matrix[np.isnan(residuals_matrix)] = np.nanmax(residuals_matrix)
        best_residuals_spectra[np.isnan(best_residuals_spectra)] = np.nanmax(best_residuals_spectra)
        kernel = int(np.sqrt(len(period_grid)))
        kernel = kernel if kernel % 2 != 0 else kernel + 1
        SR, power_raw, power, SDE_raw, SDE = ExoMoonLeastSquares.spectra(best_residuals_spectra, kernel)
        fig_transit, axs = plt.subplots(1, 1, figsize=(20, 8))
        axs.plot(period_grid, power, color='gray', alpha=1, rasterized=True, label="Flux Transit ")
        axs.set_title("Residuals")
        axs.set_xlabel('P (days)')
        axs.set_ylabel('SDE')
        plt.savefig(self.object_dir + "/SDE.png")
        plt.close(fig_transit)
        residuals_score = np.nanmax(residuals_matrix) / residuals_matrix - 1
        plt.imshow(residuals_score, aspect='auto', cmap='viridis')
        plt.colorbar()
        plt.savefig(self.object_dir + "/residuals.png")
        best_residuals_per_scenarios = np.array(best_residuals_per_scenarios)
        best_residuals_per_scenarios_plot = best_residuals_per_scenarios[np.argsort(np.array([best_residual_per_scenarios[2] for best_residual_per_scenarios in best_residuals_per_scenarios]).flatten())]
        for i in np.arange(0, 15):
            logging.info("Best residual for period %s, alpha %s: Residual->%s, Radius->%s",
                         best_residuals_per_scenarios_plot[i][0],
                         best_residuals_per_scenarios_plot[i][1], best_residuals_per_scenarios_plot[i][2],
                         best_residuals_per_scenarios_plot[i][3])
            residuals, residuals_baseline, radius, scenario_time, scenario_flux, model = self.search(
                SearchInput(best_residuals_per_scenarios_plot[i][0],
                            best_residuals_per_scenarios_plot[i][1],
                            0, 90, 0, self.min_radius, 0),
                return_lc=True)
            fig_transit, axs = plt.subplots(1, 1, figsize=(12, 12))
            axs.scatter(scenario_time, scenario_flux,
                        color='gray', alpha=0.4, rasterized=True, label="Flux")
            bin_means, bin_edges, binnumber = stats.binned_statistic(scenario_time,
                                                                     scenario_flux,
                                                                     statistic='mean', bins=25)
            bin_stds, _, _ = stats.binned_statistic(scenario_time,
                                                    scenario_flux, statistic='std', bins=25)
            bin_width = (bin_edges[1] - bin_edges[0])
            bin_centers = bin_edges[1:] - bin_width / 2
            axs.errorbar(bin_centers, bin_means, yerr=bin_stds / 2, xerr=bin_width / 2, marker='o', markersize=4,
                         color='darkorange', alpha=1, linestyle='none')
            axs.scatter(scenario_time, model,
                        color='red', alpha=1, rasterized=True, label="Model")
            axs.set_title(
                "Moon period " + str(best_residuals_per_scenarios_plot[i][0]) + " with alpha="
                + str(best_residuals_per_scenarios_plot[i][1]) + " and residual " + str(best_residuals_per_scenarios_plot[i][2]))
            axs.set_xlabel('Time')
            axs.set_ylabel('Flux')
            plt.savefig(self.object_dir + "/Residuals " + str(i) + ".png")
            plt.close(fig_transit)
        top_sde_args = np.flip(np.argsort(power))
        best_residuals_by_sde = best_residuals_per_scenarios[top_sde_args][0:15]
        for i in np.arange(0, 15):
            residuals, residuals_baseline, radius, scenario_time, scenario_flux, model = self.search(SearchInput(best_residuals_by_sde[i][0],
                                                                                      best_residuals_by_sde[i][1],
                                                                                      0, 90, 0, self.min_radius, 0),
                                                                          return_lc=True)
            fig_transit, axs = plt.subplots(1, 1, figsize=(12, 12))
            axs.scatter(scenario_time, scenario_flux,
                     color='gray', alpha=0.4, rasterized=True, label="Flux")
            bin_means, bin_edges, binnumber = stats.binned_statistic(scenario_time,
                                                                     scenario_flux,
                                                                     statistic='mean', bins=25)
            bin_stds, _, _ = stats.binned_statistic(scenario_time,
                                                    scenario_flux, statistic='std', bins=25)
            bin_width = (bin_edges[1] - bin_edges[0])
            bin_centers = bin_edges[1:] - bin_width / 2
            axs.errorbar(bin_centers, bin_means, yerr=bin_stds / 2, xerr=bin_width / 2, marker='o', markersize=4,
                         color='darkorange', alpha=1, linestyle='none')
            axs.scatter(scenario_time, model,
                     color='red', alpha=1, rasterized=True, label="Model")
            axs.set_title(
                "Moon period " + str(best_residuals_by_sde[i][0]) + " with alpha="
                + str(best_residuals_by_sde[i][1]) + " and residual " + str(best_residuals_by_sde[i][2]))
            axs.set_xlabel('Time')
            axs.set_ylabel('Flux')
            plt.savefig(self.object_dir + "/SDE " + str(i) + ".png")
            plt.close(fig_transit)


class SearchInput:
    def __init__(self, moon_period, moon_alpha, moon_ecc, moon_inc, moon_arg_periastron, min_radius, impact_param) -> None:
        self.moon_period = moon_period
        self.moon_alpha = moon_alpha
        self.moon_ecc = moon_ecc
        self.moon_inc = moon_inc
        self.moon_arg_periastron = moon_arg_periastron
        self.min_radius = min_radius
        self.impact_param = impact_param


# formatter = logging.Formatter('%(message)s')
# logger = logging.getLogger()
# while len(logger.handlers) > 0:
#     logger.handlers.pop()
# logger.setLevel(logging.INFO)
# handler = logging.StreamHandler(sys.stdout)
# handler.setLevel(logging.INFO)
# handler.setFormatter(formatter)
# logger.addHandler(handler)
# target_name = "TIC 309792357"
# planet_radius = 10.1289
# planet_period = 104.87075
# planet_t0 = 2458361.01093 - 2457000 + 0.005
# planet_duration = 6.434 / 24
# planet_inc = 89.6
# planet_ecc = 0.01
# planet_arg_periastron = 0
# planet_mass = 100
# planet_impact_param = 1
# min_depth_rms = 0.125
# max_radius = 5
# object_dir = target_name + "_EMLS"
# lc_builder = LcBuilder()
# object_info = lc_builder.build_object_info(target_name=target_name, author=None, sectors="all", file=None, cadence=120,
#                               initial_mask=None, initial_transit_mask=None, star_info=None, aperture=None,
#                               eleanor_corr_flux="pdcsap_flux", outliers_sigma=3, high_rms_enabled=False,
#                               high_rms_threshold=1.5, high_rms_bin_hours=4, smooth_enabled=False,
#                               auto_detrend_enabled=False, auto_detrend_method="cosine", auto_detrend_ratio=0.25,
#                               auto_detrend_period=None, prepare_algorithm=None, reduce_simple_oscillations=False,
#                               oscillation_snr_threshold=4, oscillation_amplitude_threshold=0.1, oscillation_ws_scale=60,
#                               oscillation_min_period=0.002, oscillation_max_period=0.2)
# if not os.path.exists(object_dir):
#     os.mkdir(object_dir)
# lc_build = lc_builder.build(object_info, object_dir)
# star_mass = lc_build.star_info.mass
# star_radius = lc_build.star_info.radius
# ab = lc_build.star_info.ld_coefficients
# times = lc_build.lc.time.value
# flux = lc_build.lc.flux.value
# flux = wotan.flatten(times, flux, method="biweight", window_length=planet_duration * 5)
# t0s = [i for i in np.arange(planet_t0, np.max(times), planet_period)]
# planet_semimajor_axis = ExoMoonLeastSquares.compute_semimajor_axis(star_mass, planet_period)
#
#
# P1 = planet_period * u.day
# a = np.cbrt((ac.G * star_mass * u.M_sun * P1 ** 2) / (4 * np.pi ** 2)).to(u.au)
# model = ellc.lc(
#     t_obs=times,
#     radius_1=(star_radius * u.R_sun).to(u.au) / a,  # star radius convert from AU to in units of a
#     radius_2=(planet_radius * u.R_earth).to(u.au) / a,
#     # convert from Rearth (equatorial) into AU and then into units of a
#     sbratio=0,
#     incl=planet_inc,
#     light_3=0,
#     t_zero=planet_t0,
#     period=planet_period,
#     a=None,
#     q=1e-6,
#     f_c=None, f_s=None,
#     ldc_1=ab, ldc_2=None,
#     gdc_1=None, gdc_2=None,
#     didt=None,
#     domdt=None,
#     rotfac_1=1, rotfac_2=1,
#     hf_1=1.5, hf_2=1.5,
#     bfac_1=None, bfac_2=None,
#     heat_1=None, heat_2=None,
#     lambda_1=None, lambda_2=None,
#     vsini_1=None, vsini_2=None,
#     t_exp=None, n_int=None,
#     grid_1='default', grid_2='default',
#     ld_1='quad', ld_2=None,
#     shape_1='sphere', shape_2='sphere',
#     spots_1=None, spots_2=None,
#     exact_grav=False, verbose=1)
#
# for t0 in t0s:
#     fig_transit, axs = plt.subplots(1, 1, figsize=(12, 12))
#     planet_duration_plot = planet_duration * 3
#     times_in_t0 = times[(times > t0 - planet_duration_plot) & (times < t0 + planet_duration_plot)]
#     if len(times_in_t0) > 20:
#         axs.scatter(times[(times > t0 - planet_duration_plot) & (times < t0 + planet_duration_plot)],
#                  flux[(times > t0 - planet_duration_plot) & (times < t0 + planet_duration_plot)],
#                  color='gray', alpha=1, rasterized=True, label="Flux")
#         axs.plot(times[(times > t0 - planet_duration_plot) & (times < t0 + planet_duration_plot)],
#                  model[(times > t0 - planet_duration_plot) & (times < t0 + planet_duration_plot)],
#                  color='red', alpha=1, rasterized=True, label="Flux")
#         axs.set_title("Planet transit at t0=" + str(t0))
#         axs.set_xlabel('Time')
#         axs.set_ylabel('Flux')
#         plt.savefig(object_dir + "/T0_" + str(t0) + ".png")
#         plt.close(fig_transit)

# emls = ExoMoonLeastSquares(object_dir, 7, star_mass, star_radius, ab, planet_radius, planet_period, planet_t0, planet_duration,
#                            planet_semimajor_axis, planet_inc, planet_ecc, planet_arg_periastron, planet_impact_param,
#                            min_depth_rms, max_radius, t0s, times, flux, period_grid_size=10000, radius_grid_size=20)
# moon_radius = 3
# moon_period = 10
# #emls.flux = emls.inject_moon(emls.time, emls.flux, t0s, planet_mass, planet_semimajor_axis, planet_ecc, moon_radius, moon_period)
# emls.planet_mass_grid = [planet_mass]
# emls.moon_inc_grid = [90]
# emls.moon_ecc_grid = [0]
# emls.moon_arg_periastron_grid = [0]
# emls.run()



"""The classes to run the plan task"""
import logging
from multiprocessing import Pool

from astroplan import moon_illumination, Constraint
import os
import shutil
import astroplan
import matplotlib
from astroplan.plots import plot_airmass
from astropy.coordinates import SkyCoord, get_moon
import astropy.units as u
from astroplan import EclipsingSystem
import pandas as pd
import numpy as np
from astroplan import (Observer, FixedTarget, AtNightConstraint, AltitudeConstraint)
import matplotlib.pyplot as plt

from astropy.time import Time, TimeDelta
from timezonefinder import TimezoneFinder
from pytz import timezone, utc


class MoonIlluminationSeparationConstraint(Constraint):
    """
    Constrain the distance between the Earth's moon and some targets.
    """

    def __init__(self, min_dist, max_dist):
        """
        Parameters
        ----------
        min_dist : `~astropy.units.Quantity`
            Minimum moon distance when moon illumination is 0
        max_dist : `~astropy.units.Quantity`
            Maximum moon distance when moon illumination is 1
        """
        self.min_dist = min_dist
        self.max_dist = max_dist

    def compute_constraint(self, times, observer, targets):
        """
        Computes the observability of the moon given a minimum distance to the moon between self.min_dist (for
        illumination = 0) and self.max_dist (for illumination = 1) by interpolating an intermediate distance from those
        two values following a linear regression.

        :param times: the times to compute the constraint for
        :param observer: the observer to compute the constraint for
        :param targets: the list of targets to compute the constraint for
        :return: the positive mask for target being observable for the given times and observer given the constraint is matched
        """
        # removed the location argument here, which causes small <1 deg
        # inaccuracies, but it is needed until astropy PR #5897 is released
        # which should be astropy 1.3.2
        moon = get_moon(times)
        # note to future editors - the order matters here
        # moon.separation(targets) is NOT the same as targets.separation(moon)
        # the former calculates the separation in the frame of the moon coord
        # which is GCRS, and that is what we want.
        moon_separation = moon.separation(targets)
        illumination = moon_illumination(times)
        min_dist = self.min_dist.value + (self.max_dist.value - self.min_dist.value) * illumination
        mask = min_dist <= moon_separation.degree
        return mask


class PlannerInput:
    def __init__(self, observatory_row, midtransit_time, observer_site, midtransit_times, ingress_egress_times,
                 constraints, moon_for_midtransit_times, moon_dist_midtransit_times, moon_phase_midtransit_times,
                 transits_since_epoch, midtransit_time_low_err, midtransit_time_up_err, low_err_delta,
                 up_err_delta, i, plan_dir, target, min_altitude, transit_fraction, baseline,
                 error_alert) -> None:
        self.observatory_row = observatory_row
        self.midtransit_time = midtransit_time
        self.observer_site = observer_site
        self.midtransit_times = midtransit_times
        self.ingress_egress_times = ingress_egress_times
        self.constraints = constraints
        self.moon_for_midtransit_times = moon_for_midtransit_times
        self.moon_dist_midtransit_times = moon_dist_midtransit_times
        self.moon_phase_midtransit_times = moon_phase_midtransit_times
        self.transits_since_epoch = transits_since_epoch
        self.midtransit_time_low_err = midtransit_time_low_err
        self.midtransit_time_up_err = midtransit_time_up_err
        self.low_err_delta = low_err_delta
        self.up_err_delta = up_err_delta
        self.i = i
        self.plan_dir = plan_dir
        self.target = target
        self.min_altitude = min_altitude
        self.transit_fraction = transit_fraction
        self.baseline = baseline
        self.error_alert = error_alert


class Planner:
    """
    Class containing the main method to generate the observation plan events.
    """

    @staticmethod
    def create_observation_observables(object_id, object_dir, ra, dec, since, name, epoch, epoch_low_err, epoch_up_err,
                                       period, period_low_err, period_up_err, duration,
                                       observatories_file, timezone, latitude, longitude, altitude,
                                       max_days, min_altitude, moon_min_dist, moon_max_dist, transit_fraction, baseline,
                                       error_alert=True, time_unit='jd', cpus=os.cpu_count() - 1):
        """
        Computes the observation windows for the given target parameters.

        :param object_id: the candidate id
        :param object_dir: the candidate directory
        :param ra: right ascension of the target
        :param dec: declination of the target
        :param since: starting plan date
        :param name: the name given to the candidate
        :param epoch: the candidate epoch
        :param epoch_low_err: the candidate epoch's lower error
        :param epoch_up_err: the candidate epoch's upper error
        :param period: the candidate period
        :param period_low_err: the candidate period's lower error
        :param period_up_err: the candidate period's upper error
        :param duration: the candidate duration
        :param observatories_file: the file containing the observatories file (csv format)
        :param timezone: the timezone of the observatory (if observatories_file=None)
        :param latitude: the latitude of the observatory (if observatories_file=None)
        :param longitude: the longitude of the observatory (if observatories_file=None)
        :param altitude: the altitude of the observatory (if observatories_file=None)
        :param max_days: the maximum number of days to compute the observables
        :param min_altitude: the minimum altitude of the target above the horizon
        :param moon_min_dist: the minimum moon distance for moon illumination = 0
        :param moon_max_dist: the minimum moon distance for moon illumination = 1
        :param transit_fraction: the minimum transit observability (0.25 for at least ingress/egress, 0.5 for ingress/egress + midtime, 1 for ingress, egress and midtime).
        :param baseline: the required baseline in hours.
        :param error_alert: whether to create the alert date to signal imprecise observations
        :param time_unit: the unit of the light curve data
        :return: the generated data and target folders observatories_df, observables_df, alert_date, plan_dir, images_dir
        :return: observatories_df containing the observatories used for the computation
        :return: observables_df containing all the observation windows that passed the plan
        :return: alert_date in case the plan reached a date where the observation uncertainty was too high
        :return: images_dir the directory where images are stored
        """
        if observatories_file is not None:
            observatories_df = pd.read_csv(observatories_file, comment='#')
        else:
            observatories_df = pd.DataFrame(columns=['name', 'tz', 'lat', 'long', 'alt'])
            observatories_df = observatories_df.append("Obs-1", timezone, latitude, longitude, altitude)
        # TODO probably convert epoch to proper JD
        primary_eclipse_time = Time(epoch, format=time_unit, scale="tdb")
        target = FixedTarget(SkyCoord(str(ra) + ' ' + str(dec), unit=(u.deg, u.deg)))
        n_transits = int(max_days // period)
        system = EclipsingSystem(primary_eclipse_time=primary_eclipse_time, orbital_period=u.Quantity(period, unit="d"),
                                 duration=u.Quantity(duration, unit="h"), name=name)
        observables_df = pd.DataFrame(columns=['observatory', 'timezone', 'start_obs', 'end_obs', 'ingress', 'egress',
                                               'midtime', "midtime_up_err_h", "midtime_low_err_h", 'twilight_evening',
                                               'twilight_morning', 'observable', 'moon_phase', 'moon_dist'])
        plan_dir = object_dir + "/plan"
        images_dir = plan_dir + "/images"
        if not os.path.exists(plan_dir):
            os.mkdir(plan_dir)
        if os.path.exists(images_dir):
            shutil.rmtree(images_dir, ignore_errors=True)
        os.mkdir(images_dir)
        alert_date = None
        planner_inputs = []
        for index, observatory_row in observatories_df.iterrows():
            observer_site = Observer(latitude=observatory_row["lat"], longitude=observatory_row["lon"],
                                     elevation=u.Quantity(observatory_row["alt"], unit="m"))
            midtransit_times = system.next_primary_eclipse_time(since, n_eclipses=n_transits)
            ingress_egress_times = system.next_primary_ingress_egress_time(since, n_eclipses=n_transits)
            constraints = [AtNightConstraint.twilight_nautical(), AltitudeConstraint(min=min_altitude * u.deg),
                           MoonIlluminationSeparationConstraint(min_dist=moon_min_dist * u.deg,
                                                                max_dist=moon_max_dist * u.deg)]
            moon_for_midtransit_times = get_moon(midtransit_times)
            moon_dist_midtransit_times = moon_for_midtransit_times.separation(SkyCoord(ra, dec, unit="deg"))
            moon_phase_midtransit_times = np.round(astroplan.moon_illumination(midtransit_times), 2)
            transits_since_epoch = np.round((midtransit_times - primary_eclipse_time).jd / period)
            midtransit_time_low_err = np.round(
                (((transits_since_epoch * period_low_err) ** 2 + epoch_low_err ** 2) ** (1 / 2)) * 24, 2)
            midtransit_time_up_err = np.round(
                (((transits_since_epoch * period_up_err) ** 2 + epoch_up_err ** 2) ** (1 / 2)) * 24, 2)
            low_err_delta = TimeDelta(midtransit_time_low_err * 3600, format='sec')
            up_err_delta = TimeDelta(midtransit_time_up_err * 3600, format='sec')
            i = 0
            for midtransit_time in midtransit_times:
                planner_inputs = planner_inputs + [
                    PlannerInput(observatory_row, midtransit_time, observer_site, midtransit_times,
                                 ingress_egress_times,
                                 constraints, moon_for_midtransit_times, moon_dist_midtransit_times,
                                 moon_phase_midtransit_times,
                                 transits_since_epoch, midtransit_time_low_err, midtransit_time_up_err, low_err_delta,
                                 up_err_delta, i, plan_dir, target, min_altitude, transit_fraction, baseline,
                                 error_alert)]
                i = i + 1
        with Pool(processes=cpus) as pool:
            alert_dates_and_observables = pool.map(Planner.plan_event, planner_inputs)
        alert_dates = [item[0] for item in alert_dates_and_observables]
        alert_date = alert_dates[0] if len(alert_dates) > 0 else None
        for each_alert_date in alert_dates:
            if each_alert_date is not None and each_alert_date < alert_date:
                alert_date = each_alert_date
        observables = [item[1] for item in alert_dates_and_observables]
        for observable in observables:
            if observable is not None:
                observables_df = observables_df.append(observable, ignore_index=True)
        observables_df = observables_df.sort_values(["midtime", "observatory"], ascending=True)
        observables_df.to_csv(plan_dir + "/observation_plan.csv", index=False)
        print("Observation plan created in directory: " + object_dir)
        return observatories_df, observables_df, alert_date, plan_dir, images_dir

    @staticmethod
    def plan_event(planner_input):
        try:
            twilight_evening = planner_input.observer_site.twilight_evening_nautical(planner_input.midtransit_time)
            twilight_morning = planner_input.observer_site.twilight_morning_nautical(planner_input.midtransit_time)
            ingress = planner_input.ingress_egress_times[planner_input.i][0]
            egress = planner_input.ingress_egress_times[planner_input.i][1]
            lowest_ingress = ingress - planner_input.low_err_delta[planner_input.i]
            highest_egress = egress + planner_input.up_err_delta[planner_input.i]
            if planner_input.error_alert and (highest_egress - lowest_ingress).jd > 0.33:
                alert_date = planner_input.midtransit_time \
                    if (planner_input.alert_date is None) or \
                       (planner_input.alert_date is not None and planner_input.alert_date >= planner_input.midtransit_time) \
                    else planner_input.alert_date
                return (alert_date, None)
            else:
                baseline_low = lowest_ingress - planner_input.baseline * u.hour
                baseline_up = highest_egress + planner_input.baseline * u.hour
                transit_times = baseline_low + (baseline_up - baseline_low) * np.linspace(0, 1, 100)
                observable_transit_times = astroplan.is_event_observable(planner_input.constraints,
                                                                         planner_input.observer_site,
                                                                         planner_input.target,
                                                                         times=transit_times)[0]
                observable_transit_times_true = np.argwhere(observable_transit_times)
                observable = len(observable_transit_times_true) / 100
                if observable < planner_input.transit_fraction:
                    return (None, None)
                start_obs = transit_times[observable_transit_times_true[0]][0]
                end_obs = transit_times[observable_transit_times_true[len(observable_transit_times_true) - 1]][0]
                start_plot = baseline_low
                end_plot = baseline_up
                # TODO check whether twilight evening happens before twilight morning, if not, the check is different
                if twilight_evening > start_obs:
                    start_obs = twilight_evening
                if twilight_morning < end_obs:
                    end_obs = twilight_morning
            moon_dist = round(planner_input.moon_dist_midtransit_times[planner_input.i].degree)
            moon_phase = planner_input.moon_phase_midtransit_times[planner_input.i]
            # TODO get is_event_observable for several parts of the transit (ideally each 5 mins) to get the proper observable percent. Also with baseline
            if planner_input.observatory_row["tz"] is not None and not np.isnan(planner_input.observatory_row["tz"]):
                observer_timezone = planner_input.observatory_row["tz"]
            else:
                observer_timezone = Planner.get_offset(planner_input.observatory_row["lat"], planner_input.observatory_row["lon"],
                                                       planner_input.midtransit_time.datetime)
            observable_dict = {"observatory": planner_input.observatory_row["name"],
                                                    "timezone": observer_timezone, "ingress": ingress.isot,
                                                    "start_obs": start_obs.isot, "end_obs": end_obs.isot,
                                                    "egress": egress.isot, "midtime": planner_input.midtransit_time.isot,
                                                    "midtime_up_err_h":
                                                        str(int(planner_input.midtransit_time_up_err[planner_input.i] // 1)) + ":" +
                                                        str(int(planner_input.midtransit_time_up_err[planner_input.i] % 1 * 60)).zfill(2),
                                                    "midtime_low_err_h":
                                                        str(int(planner_input.midtransit_time_low_err[planner_input.i] // 1)) + ":" +
                                                        str(int(planner_input.midtransit_time_low_err[planner_input.i] % 1 * 60)).zfill(2),
                                                    "twilight_evening": twilight_evening.isot,
                                                    "twilight_morning": twilight_morning.isot,
                                                    "observable": observable, "moon_phase": moon_phase,
                                                    "moon_dist": moon_dist}
            plot_time = start_plot + (end_plot - start_plot) * np.linspace(0, 1, 100)
            plt.tick_params(labelsize=6)
            airmass_ax = plot_airmass(planner_input.target, planner_input.observer_site, plot_time, brightness_shading=False,
                                      altitude_yaxis=True)
            airmass_ax.axvspan(twilight_morning.plot_date, end_plot.plot_date, color='white')
            airmass_ax.axvspan(start_plot.plot_date, twilight_evening.plot_date, color='white')
            airmass_ax.axvspan(twilight_evening.plot_date, twilight_morning.plot_date, color='gray')
            airmass_ax.axhspan(1. / np.cos(np.radians(90 - planner_input.min_altitude)), 5.0, color='green')
            airmass_ax.get_figure().gca().set_title("")
            airmass_ax.get_figure().gca().set_xlabel("")
            airmass_ax.get_figure().gca().set_ylabel("")
            airmass_ax.set_xlabel("")
            airmass_ax.set_ylabel("")
            xticks = []
            xticks_labels = []
            xticks.append(start_obs.plot_date)
            hour_min_sec_arr = start_obs.isot.split("T")[1].split(":")
            xticks_labels.append("T1_" + hour_min_sec_arr[0] + ":" + hour_min_sec_arr[1])
            plt.axvline(x=start_obs.plot_date, color="violet")
            xticks.append(end_obs.plot_date)
            hour_min_sec_arr = end_obs.isot.split("T")[1].split(":")
            xticks_labels.append("T1_" + hour_min_sec_arr[0] + ":" + hour_min_sec_arr[1])
            plt.axvline(x=end_obs.plot_date, color="violet")
            if start_plot < lowest_ingress < end_plot:
                xticks.append(lowest_ingress.plot_date)
                hour_min_sec_arr = lowest_ingress.isot.split("T")[1].split(":")
                xticks_labels.append("T1_" + hour_min_sec_arr[0] + ":" + hour_min_sec_arr[1])
                plt.axvline(x=lowest_ingress.plot_date, color="red")
            if start_plot < ingress < end_plot:
                xticks.append(ingress.plot_date)
                hour_min_sec_arr = ingress.isot.split("T")[1].split(":")
                xticks_labels.append("T1_" + hour_min_sec_arr[0] + ":" + hour_min_sec_arr[1])
                plt.axvline(x=ingress.plot_date, color="orange")
            if start_plot < planner_input.midtransit_time < end_plot:
                xticks.append(planner_input.midtransit_time.plot_date)
                hour_min_sec_arr = planner_input.midtransit_time.isot.split("T")[1].split(":")
                xticks_labels.append("T0_" + hour_min_sec_arr[0] + ":" + hour_min_sec_arr[1])
                plt.axvline(x=planner_input.midtransit_time.plot_date, color="black")
            if start_plot < egress < end_plot:
                xticks.append(egress.plot_date)
                hour_min_sec_arr = egress.isot.split("T")[1].split(":")
                xticks_labels.append("T4_" + hour_min_sec_arr[0] + ":" + hour_min_sec_arr[1])
                plt.axvline(x=egress.plot_date, color="orange")
            if start_plot < highest_egress < end_plot:
                xticks.append(highest_egress.plot_date)
                hour_min_sec_arr = highest_egress.isot.split("T")[1].split(":")
                xticks_labels.append("T4_" + hour_min_sec_arr[0] + ":" + hour_min_sec_arr[1])
                plt.axvline(x=highest_egress.plot_date, color="red")
            airmass_ax.xaxis.set_tick_params(labelsize=5)
            airmass_ax.set_xticks([])
            airmass_ax.set_xticklabels([])
            degrees_ax = Planner.get_twin(airmass_ax)
            degrees_ax.yaxis.set_tick_params(labelsize=6)
            degrees_ax.set_yticks([1., 1.55572383, 2.])
            degrees_ax.set_yticklabels([90, 50, 30])
            fig = matplotlib.pyplot.gcf()
            fig.set_size_inches(1.25, 0.75)
            plt.savefig(
                planner_input.plan_dir + "/images/" + planner_input.observatory_row["name"] + "_" +
                str(planner_input.midtransit_time.isot)[:-4] + ".png",
                bbox_inches='tight')
            plt.close()
            return (None, observable_dict)
        except Exception as e:
            logging.exception("Error when planning one observation event.")

    @staticmethod
    def get_twin(ax):
        """
        Retrieves a twin Y axis for a given matplotlib axis. This is useful when we have two axes one placed at each side
        of the plot.

        :param ax: the known matplotlib axis.
        :return: the twin axis.
        """
        for other_ax in ax.figure.axes:
            if other_ax is ax:
                continue
            if other_ax.bbox.bounds == ax.bbox.bounds:
                return other_ax
        return None

    @staticmethod
    def get_offset(lat, lng, datetime):
        """
        Returns a location's time zone offset from UTC in minutes.

        :param lat: geographical latitude
        :param lng: geographical longitude
        :param datetime: the UTC time
        """
        tf = TimezoneFinder()
        tz_target = timezone(tf.certain_timezone_at(lng=lng, lat=lat))
        if tz_target is None:
            return None
        today_target = tz_target.localize(datetime)
        today_utc = utc.localize(datetime)
        return (today_utc - today_target).total_seconds() / 3600



from abc import ABC, abstractmethod


class SignalSelector(ABC):
    """
    Base class to perform a selection of the best candidate signal among a list of them.
    """
    def __init__(self):
        pass

    @abstractmethod
    def select(self, id_run, sherlock_target, star_info, transits_min_count, time, lcs, transit_results, wl, cadence):
        """
        Selects the best signal given the implementation of this method.

        :param transit_result_detrends: a list of transit results to be explored.
        :param snr_min: the minimum snr to allow a signal to be considered good enough.
        :param sde_min: the minimum sde to allow a signal to be considered good enough.
        :param detrend_method: the detrend method used to obtain the candidate transit results
        :param wl: the window length used by the detrend method used.
        :return: a SignalSelection with the parameters that characterize the best selection.
        """
        pass

class SignalSelection:
    def __init__(self, score, curve_index, transit_result):
        self.score = score
        self.curve_index = curve_index
        self.transit_result = transit_result

    def get_message(self):
        curve_name = "PDCSAP_FLUX" if self.curve_index == 0 else str(self.curve_index - 1)
        return "Chosen signal with BASIC algorithm --> NAME: " + curve_name + \
               "\tPeriod:" + str(self.transit_result.period) + \
               "\tSNR: " + str(self.transit_result.snr) + \
               "\tSDE: " + str(self.transit_result.sde) + \
               "\tFAP: " + str(self.transit_result.fap) + \
               "\tBORDER_SCORE: " + str(self.transit_result.border_score)



import copy
import logging
import shutil
import traceback
import numpy as np
import matplotlib.pyplot as plt
import yaml
from astroquery.mast import TesscutClass
from lcbuilder.lcbuilder_class import LcBuilder
from lightkurve import TessLightCurve, KeplerLightCurve
import pandas as pd
import os
import ast
import triceratops.triceratops as tr
from watson.watson import Watson

from sherlockpipe.validation.validation_report import ValidationReport
from sherlockpipe.loading.tool_with_candidate import ToolWithCandidate


class Validator(ToolWithCandidate):
    """
    This class intends to provide a statistical validation tool for SHERLOCK Candidates.
    """
    def __init__(self, object_dir, validation_dir, is_candidate_from_search, candidates_df):
        super().__init__(is_candidate_from_search, candidates_df)
        self.object_dir = os.getcwd() if object_dir is None else object_dir
        self.data_dir = validation_dir

    def validate(self, candidate, star, cpus, contrast_curve_file, bins=100, scenarios=5, sigma_mode="flux_err"):
        """
        Runs the validation storing a PDF report with the results together to some csvs for reproducibility.

        :param candidate: a candidate dataframe containing TICID, period, duration, t0, transits, depth, rp_rs, number, curve and sectors data
        :param star: the star dataframe
        :param cpus: the number of cpus to be used
        :param contrast_curve_file: the auxiliary contrast curve file to give more information to the validation engine
        :param bins: the number of bins to resize the light curve
        :param scenarios: the number of scenarios to compute the validation and get the average
        :param sigma_mode: whether to compute the sigma for the validation from the 'flux_err' or the 'binning'
        """
        object_id = candidate["id"]
        period = candidate.loc[candidate['id'] == object_id]['period'].iloc[0]
        duration = candidate.loc[candidate['id'] == object_id]['duration'].iloc[0]
        t0 = candidate.loc[candidate['id'] == object_id]['t0'].iloc[0]
        transit_depth = candidate.loc[candidate['id'] == object_id]['depth'].iloc[0]
        run = int(candidate.loc[candidate['id'] == object_id]['number'].iloc[0])
        curve = int(candidate.loc[candidate['id'] == object_id]['curve'].iloc[0])
        rp_rstar = candidate.loc[candidate['id'] == object_id]['rp_rs'].iloc[0]
        a_rstar = candidate.loc[candidate['id'] == object_id]['a'].iloc[0] / star["R_star"]
        logging.info("------------------")
        logging.info("Candidate info")
        logging.info("------------------")
        logging.info("Period (d): %.2f", period)
        logging.info("Epoch (d): %.2f", t0)
        logging.info("Duration (min): %.2f", duration)
        logging.info("Depth (ppt): %.2f", transit_depth)
        logging.info("Run: %.0f", run)
        logging.info("Detrend curve: %.0f", curve)
        logging.info("Contrast curve file %s", contrast_curve_file)
        lc_file = "/lc_" + str(curve) + ".csv"
        lc_file = self.object_dir + lc_file
        try:
            sectors_in = ast.literal_eval(str((candidate.loc[candidate['id'] == object_id]['sectors']).values[0]))
            if (type(sectors_in) == int) or (type(sectors_in) == float):
                sectors = [sectors_in]
            else:
                sectors = list(sectors_in)
        except:
            sectors = [0]
        object_id = object_id.iloc[0]
        try:
            self.execute_triceratops(cpus, self.data_dir, object_id, sectors, lc_file, transit_depth,
                                          period, t0, duration, rp_rstar, a_rstar, bins, scenarios, sigma_mode,
                                          contrast_curve_file, run, star_df=star)
            report = ValidationReport(self.data_dir + '/triceratops', "statistical_validation_report.pdf", object_id, star['ra'], star['dec'],
                                       t0, period, duration, transit_depth, star["v"], star["j"], star["h"],
                                       star["k"])
            report.create_report()
            for file in os.listdir(self.data_dir + '/triceratops'):
                if "validation" not in file and (".pdf" in file or ".png" in file):
                    os.remove(self.data_dir + '/triceratops/' + file)
        except Exception as e:
            traceback.print_exc()
        # try:
        #     self.execute_vespa(cpus, validation_dir, object_id, sectors, lc_file, transit_depth, period, t0, duration, rprs)
        # except Exception as e:
        #     traceback.print_exc()

    def execute_triceratops(self, cpus, indir, object_id, sectors, lc_file, transit_depth, period, t0,
                            transit_duration, rp_rstar, a_rstar, bins, scenarios, sigma_mode, contrast_curve_file,
                            run, star_df):
        """ Calculates probabilities of the signal being caused by any of the following astrophysical sources:
        TP No unresolved companion. Transiting planet with Porb around target star. (i, Rp)
        EB No unresolved companion. Eclipsing binary with Porb around target star. (i, qshort)
        EBx2P No unresolved companion. Eclipsing binary with 2 × Porb around target star. (i, qshort)
        PTP Unresolved bound companion. Transiting planet with Porb around primary star. (i, Rp, qlong)
        PEB Unresolved bound companion. Eclipsing binary with Porb around primary star. (i, qshort, qlong)
        PEBx2P Unresolved bound companion. Eclipsing binary with 2 × Porb around primary star. (i, qshort, qlong)
        STP Unresolved bound companion. Transiting planet with Porb around secondary star. (i, Rp, qlong)
        SEB Unresolved bound companion. Eclipsing binary with Porb around secondary star. (i, qshort, qlong)
        SEBx2P Unresolved bound companion. Eclipsing binary with 2 × Porb around secondary star. (i, qshort, qlong)
        DTP Unresolved background star. Transiting planet with Porb around target star. (i, Rp, simulated star)
        DEB Unresolved background star. Eclipsing binary with Porb around target star. (i, qshort, simulated star)
        DEBx2P Unresolved background star. Eclipsing binary with 2 × Porb around target star. (i, qshort, simulated star)
        BTP Unresolved background star. Transiting planet with Porb around background star. (i, Rp, simulated star)
        BEB Unresolved background star. Eclipsing binary with Porb around background star. (i, qshort, simulated star)
        BEBx2P Unresolved background star. Eclipsing binary with 2 × Porb around background star. (i, qshort, simulated star)
        NTP No unresolved companion. Transiting planet with Porb around nearby star. (i, Rp)
        NEB No unresolved companion. Eclipsing binary with Porb around nearby star. (i, qshort)
        NEBx2P No unresolved companion. Eclipsing binary with 2 × Porb around nearby star. (i, qshort)
        FPP = 1 - (TP + PTP + DTP)
        NFPP = NTP + NEB + NEBx2P
        Giacalone & Dressing (2020) define validated planets as TOIs with NFPP < 10−3 and FPP < 0.015 (or FPP ≤ 0.01,
        when rounding to the nearest percent)

        :param cpus: number of cpus to be used
        :param indir: root directory to store the results
        :param id_int: the object id for which the analysis will be run
        :param sectors: the sectors of the tic
        :param lc_file: the light curve source file
        :param transit_depth: the depth of the transit signal (ppts)
        :param period: the period of the transit signal (days)
        :param t0: the t0 of the transit signal (days)
        :param transit_duration: the duration of the transit signal (minutes)
        :param rp_rstar: radius of planet divided by radius of star
        :param a_rstar: semimajor axis divided by radius of star
        :param bins: the number of bins to average the folded curve
        :param scenarios: the number of scenarios to validate
        :param sigma_mode: the way to calculate the sigma for the validation ['flux_err' | 'binning']
        :param contrast_curve_file: the auxiliary contrast curve file to give more information to the validation engine
        :param run: the search run where the candidate was spotted
        :param star_df: the star dataframe containing the host parameters
        :return str: the directory where the results are stored
        """
        save_dir = indir + "/triceratops"
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir, ignore_errors=True)
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        duration = transit_duration / 60 / 24
        logging.info("----------------------")
        logging.info("Validation procedures")
        logging.info("----------------------")
        logging.info("Pre-processing sectors")
        mission, mission_prefix, id_int = LcBuilder().parse_object_info(object_id)
        sectors = np.array(sectors)
        if mission == "TESS":
            sectors_cut = TesscutClass().get_sectors(objectname="TIC " + str(id_int))
            sectors_cut = np.array([sector_row["sector"] for sector_row in sectors_cut])
            if len(sectors) != len(sectors_cut):
                logging.warning("WARN: Some sectors were not found in TESSCUT")
                logging.warning("WARN: Sherlock sectors were: " + str(sectors))
                logging.warning("WARN: TESSCUT sectors were: " + str(sectors_cut))
            sectors = np.intersect1d(sectors, sectors_cut)
            if len(sectors) == 0:
                logging.warning("There are no available sectors to be validated, skipping TRICERATOPS.")
                return save_dir, None, None
        logging.info("Will execute validation for sectors: " + str(sectors))
        lc = pd.read_csv(lc_file, header=0)
        time, flux, flux_err = lc["#time"].values, lc["flux"].values, lc["flux_err"].values
        time, flux, flux_err = self.mask_previous_candidates(time, flux, flux_err, run)
        if contrast_curve_file is not None:
            logging.info("Reading contrast curve %s", contrast_curve_file)
            plt.clf()
            cc = pd.read_csv(contrast_curve_file, header=None)
            sep, dmag = cc[0].values, cc[1].values
            plt.plot(sep, dmag, 'k-')
            plt.ylim(9, 0)
            plt.ylabel("$\\Delta K_s$", fontsize=20)
            plt.xlabel("separation ('')", fontsize=20)
            plt.savefig(save_dir + "/contrast_curve.png")
            plt.clf()
            shutil.copy(contrast_curve_file, save_dir + "/cc_" + os.path.basename(contrast_curve_file))
        logging.info("Preparing validation light curve for target")
        lc_len = len(time)
        zeros_lc = np.zeros(lc_len)
        depth = transit_depth / 1000
        if mission == "TESS":
            lc = TessLightCurve(time=time, flux=flux, flux_err=flux_err, quality=zeros_lc)
        else:
            lc = KeplerLightCurve(time=time, flux=flux, flux_err=flux_err, quality=zeros_lc)
        lc.extra_columns = []
        fig, axs = plt.subplots(1, 1, figsize=(8, 4), constrained_layout=True)
        axs, bin_centers, bin_means, bin_stds, snr = \
            Watson.compute_phased_values_and_fill_plot(object_id, axs, lc, period, t0, depth, duration, rp_rstar,
                                                       a_rstar, bins=bins,
                                                       bin_err_mode='bin' if sigma_mode == 'binning' else 'flux_err')
        bin_centers = (bin_centers - 0.5) * period
        plt.savefig(save_dir + "/folded_curve.png")
        plt.clf()
        logging.info("Sigma mode is %s", sigma_mode)
        sigma = np.nanmean(bin_stds)
        logging.info("Flux err (ppm) = %s", sigma * 1000000)
        logging.info("Acquiring triceratops target")
        target = tr.target(ID=id_int, mission=mission, sectors=sectors)
        # TODO allow user input apertures
        logging.info("Reading apertures from directory")
        apertures = yaml.load(open(self.object_dir + "/apertures.yaml"), yaml.SafeLoader)
        apertures = apertures["sectors"]
        valid_apertures = {}
        for sector, aperture in apertures.items():
            if sector in sectors:
                valid_apertures[sector] = aperture
                target.plot_field(save=True, fname=save_dir + "/field_S" + str(sector), sector=sector,
                                  ap_pixels=aperture)
        valid_apertures = np.array([aperture for sector, aperture in valid_apertures.items()], dtype=object)
        logging.info("Calculating validation closest stars depths")
        target.calc_depths(depth, valid_apertures)
        target.stars.to_csv(save_dir + "/stars.csv", index=False)
        if np.isnan(target.stars.loc[0, 'mass']):
            target.stars.loc[0, 'mass'] = star_df['M_star']
        if np.isnan(target.stars.loc[0, 'rad']):
            target.stars.loc[0, 'rad'] = star_df['R_star']
        if np.isnan(target.stars.loc[0, 'Teff']):
            target.stars.loc[0, 'Teff'] = star_df['Teff']
        logging.info("Preparing validation processes inputs")
        input_n_times = [ValidatorInput(save_dir, copy.deepcopy(target), bin_centers, bin_means, sigma, period, depth,
                                        valid_apertures, value, contrast_curve_file)
                         for value in range(0, scenarios)]
        logging.info("Start validation processes")
        #TODO fix usage of cpus returning same value for all executions
        validation_results = []
        for i in range(0, scenarios):
            validation_results.append(TriceratopsThreadValidator.validate(input_n_times[i]))
        # with Pool(processes=1) as pool:
        #     validation_results = pool.map(TriceratopsThreadValidator.validate, input_n_times)
        logging.info("Finished validation processes")
        fpp_sum = 0
        fpp2_sum = 0
        fpp3_sum = 0
        nfpp_sum = 0
        probs_total_df = None
        scenarios_num = len(validation_results[0][4])
        star_num = np.zeros((5, scenarios_num))
        u1 = np.zeros((5, scenarios_num))
        u2 = np.zeros((5, scenarios_num))
        fluxratio_EB = np.zeros((5, scenarios_num))
        fluxratio_comp = np.zeros((5, scenarios_num))
        target = input_n_times[0].target
        target.star_num = np.zeros(scenarios_num)
        target.u1 = np.zeros(scenarios_num)
        target.u2 = np.zeros(scenarios_num)
        target.fluxratio_EB = np.zeros(scenarios_num)
        target.fluxratio_comp = np.zeros(scenarios_num)
        logging.info("Computing final probabilities from the %s scenarios", scenarios)
        fpps = [validation_result[0] for validation_result in validation_results]
        nfpps = [validation_result[1] for validation_result in validation_results]
        fpp_err = np.std(fpps)
        nfpp_err = np.std(nfpps)
        i = 0
        with open(save_dir + "/validation.csv", 'w') as the_file:
            the_file.write("scenario,FPP,NFPP,FPP2,FPP3+\n")
            for fpp, nfpp, fpp2, fpp3, probs_df, star_num_arr, u1_arr, u2_arr, fluxratio_EB_arr, fluxratio_comp_arr \
                    in validation_results:
                if probs_total_df is None:
                    probs_total_df = probs_df
                else:
                    probs_total_df = pd.concat((probs_total_df, probs_df))
                fpp_sum = fpp_sum + fpp
                fpp2_sum = fpp2_sum + fpp2
                fpp3_sum = fpp3_sum + fpp3
                nfpp_sum = nfpp_sum + nfpp
                star_num[i] = star_num_arr
                u1[i] = u1_arr
                u2[i] = u2_arr
                fluxratio_EB[i] = fluxratio_EB_arr
                fluxratio_comp[i] = fluxratio_comp_arr
                the_file.write(str(i) + "," + str(fpp) + "," + str(nfpp) + "," + str(fpp2) + "," + str(fpp3) + "\n")
                i = i + 1
            for i in range(0, scenarios_num):
                target.star_num[i] = np.mean(star_num[:, i])
                target.u1[i] = np.mean(u1[:, i])
                target.u2[i] = np.mean(u2[:, i])
                target.fluxratio_EB[i] = np.mean(fluxratio_EB[:, i])
                target.fluxratio_comp[i] = np.mean(fluxratio_comp[:, i])
            fpp_sum = fpp_sum / scenarios
            nfpp_sum = nfpp_sum / scenarios
            fpp2_sum = fpp2_sum / scenarios
            fpp3_sum = fpp3_sum / scenarios
            the_file.write("MEAN" + "," + str(fpp_sum) + "," + str(nfpp_sum) + "," + str(fpp2_sum) + "," +
                           str(fpp3_sum))
        logging.info("Plotting mean scenario outputs")
        Validator.plot_triceratops_output(fpp_sum, nfpp_sum, fpp_err, nfpp_err, save_dir)
        probs_total_df = probs_total_df.groupby(["ID", "scenario"], as_index=False).mean()
        target.probs = probs_total_df
        target.plot_fits(save=True, fname=save_dir + "/scenario_fits", time=bin_centers, flux_0=bin_means,
                         flux_err_0=sigma)
        probs_total_df["scenario"] = pd.Categorical(probs_total_df["scenario"], ["TP", "EB", "EBx2P", "PTP", "PEB",
                                                                                 "PEBx2P", "STP", "SEB", "SEBx2P",
                                                                                 "DTP", "DEB", "DEBx2P", "BTP", "BEB",
                                                                                 "BEBx2P", "NTP", "NEB", "NEBx2P"])
        probs_total_df = probs_total_df.sort_values("scenario")
        probs_total_df.to_csv(save_dir + "/validation_scenarios.csv", index=False)
        logging.info("---------------------------------")
        logging.info("Final probabilities computed")
        logging.info("---------------------------------")
        logging.info("FPP=%s", fpp_sum)
        logging.info("NFPP=%s", nfpp_sum)
        logging.info("FPP2(Lissauer et al, 2012)=%s", fpp2_sum)
        logging.info("FPP3+(Lissauer et al, 2012)=%s", fpp3_sum)
        return save_dir

    @staticmethod
    def plot_triceratops_output(fpp, nfpp, fpp_err, nfpp_err, target_dir):
        """
        Given the TRICERATOPS informed FPP and NFPP, creates a plot with the information and the FP and Likely Planet
        thresholds

        :param fpp: the False Positive Probability
        :param nfpp: the Nearby False Positive Probability
        :param fpp_err: the False Positive Probability Error
        :param nfpp_err: the Nearby False Positive Probability Error
        :param target_dir: directory to store the plot
        """
        min_fpp = 0.000001
        likely = (0.5, 0.001)
        likely_nfpp = 0.1
        fig, axs = plt.subplots(1, 1, figsize=(8, 8), constrained_layout=True)
        nfpp = nfpp if nfpp > min_fpp else min_fpp
        fpp = fpp if fpp > min_fpp else min_fpp
        #axs.set_xlim([0, 1])
        #axs.set_ylim([0, 1])
        axs.set_yscale('log', base=10)
        #axs.set_yscale('log', base=10)
        axs.set_title("FPP / NFPP Map", fontsize=25)
        axs.set_xlabel('FPP', fontsize=25)
        axs.set_ylabel('NFPP', fontsize=25)
        axs.axhspan(0, likely[1], 0, likely[0], color="lightgreen", label="Likely Planet")
        axs.axhspan(likely_nfpp, 1, color="gainsboro", label="Likely NFP")
        axs.errorbar(fpp, nfpp, xerr=fpp_err, yerr=nfpp_err, marker="o", markersize=15, markerfacecolor="blue",
                     ecolor="cyan")
        axs.plot(min_fpp, min_fpp, markersize=0)
        #axs.legend(loc='upper right', fontsize=25)
        axs.text(0.77, 0.115, "Likely NFP", fontsize=20)
        axs.text(0.23, 0.00057, "Likely Planet", fontsize=20)
        axs.tick_params(axis='both', which='major', labelsize=15)
        axs.set_xlim([0, 1])
        fig.savefig(target_dir + "/triceratops_map.png")
        fig.clf()

    @staticmethod
    def probs_without_scenarios(csv_file, no_scenarios):
        """
        Helper method to re-compute the probabilities removing the given scenarios

        :param csv_file: the csv file with all the scenarios probabilities
        :param no_scenarios: the scenarios to be removed from the calculation of probabilities
        :return: the new fpp, nfpp, fpp2, fpp3 and the filtered scenarios dataframe
        """
        scenarios_df = pd.read_csv(csv_file)
        scenarios_prob = scenarios_df['prob'].sum()
        filtered_scenarios_df = scenarios_df.loc[~scenarios_df['scenario'].isin(no_scenarios)]
        filtered_prob = filtered_scenarios_df['prob'].sum()
        filtered_scenarios_df['prob'] = filtered_scenarios_df['prob'] * scenarios_prob / filtered_prob
        filtered_prob_sum = filtered_scenarios_df['prob'].sum()
        filtered_scenarios_df['prob'] = filtered_scenarios_df['prob'] / filtered_prob_sum
        fpp = 1 - filtered_scenarios_df[filtered_scenarios_df['scenario'].isin(['TP', 'PTP', 'DTP'])]['prob'].sum()
        fpp2 = 1 - 25 * (1 - fpp) / (25 * (1 - fpp) + fpp)
        fpp3 = 1 - 50 * (1 - fpp) / (50 * (1 - fpp) + fpp)
        nfpp = filtered_scenarios_df[filtered_scenarios_df['scenario'].isin(['NTP', 'NEB', 'NEBx2P'])]['prob'].sum()
        print("Filtered scenarios " + str(no_scenarios) + " for file " + csv_file)
        print("FPP: " + str(fpp) + "   FPP2: " + str(fpp2) + "   FPP3+: " + str(fpp3) + "   NFPP: " + str(nfpp))
        return fpp, nfpp, fpp2, fpp3, filtered_scenarios_df


    # def execute_vespa(self, cpus, indir, object_id, sectors, lc_file, transit_depth, period, epoch, duration, rprs):
    #     vespa_dir = indir + "/vespa/"
    #     if os.path.exists(vespa_dir):
    #         shutil.rmtree(vespa_dir, ignore_errors=True)
    #     os.mkdir(vespa_dir)
    #     lc = pd.read_csv(lc_file, header=0)
    #     time, flux, flux_err = lc["#time"].values, lc["flux"].values, lc["flux_err"].values
    #     lc_len = len(time)
    #     zeros_lc = np.zeros(lc_len)
    #     logging.info("Preparing validation light curve for target")
    #     lc = TessLightCurve(time=time, flux=flux, flux_err=flux_err, quality=zeros_lc)
    #     lc.extra_columns = []
    #     time_float = lc.time.value
    #     cadence_array = np.diff(time_float)
    #     cadence_array = cadence_array[~np.isnan(cadence_array)]
    #     cadence_array = cadence_array[cadence_array > 0]
    #     cadence_days = np.nanmedian(cadence_array)
    #     lc = lc.fold(period=period, epoch_time=epoch, normalize_phase=True)
    #     folded_plot_range = duration / 60 / 24 / 2 / period * 5
    #     inner_folded_range_args = np.where(
    #         (0 - folded_plot_range < lc.time.value) & (lc.time.value < 0 + folded_plot_range))
    #     lc = lc[inner_folded_range_args]
    #     lc.time = lc.time * period
    #     bin_means, bin_edges, binnumber = stats.binned_statistic(lc.time.value, lc.flux.value, statistic='mean',
    #                                                              bins=500)
    #     bin_means_err, bin_edges_err, binnumber_err = stats.binned_statistic(lc.time.value, lc.flux_err.value, statistic='mean',
    #                                                              bins=500)
    #     bin_width = (bin_edges[1] - bin_edges[0])
    #     bin_centers = bin_edges[1:] - bin_width / 2
    #     lc.plot()
    #     plt.title("Target " + str(object_id))
    #     plt.savefig(vespa_dir + "/folded_curve.png")
    #     plt.plot(bin_centers, bin_means)
    #     plt.title("Target " + str(object_id))
    #     plt.xlabel("Time")
    #     plt.ylabel("Flux")
    #     plt.savefig(vespa_dir + "/folded_curve_binned.png")
    #     lc_df = pandas.DataFrame(columns=['days_from_midtransit', 'flux', 'flux_err'])
    #     lc_df['days_from_midtransit'] = bin_centers
    #     lc_df['flux'] = bin_means
    #     lc_df['flux_err'] = bin_means_err
    #     lc_df.to_csv(vespa_dir + "/lc.csv", index=False)
    #     star_df = pd.read_csv(self.object_dir + "/params_star.csv")
    #     with open(vespa_dir + "fpp.ini", 'w+') as f:
    #         f.write("name = " + str(object_id) + "\n")
    #         f.write("ra = " + str(star_df.at[0, "ra"]) + "\n")
    #         f.write("dec = " + str(star_df.at[0, "dec"]) + "\n")
    #         f.write("period = " + str(period) + "\n")
    #         f.write("rprs = " + str(rprs) + "\n")
    #         # TODO rewrite lc file to match vespa expected
    #         f.write("photfile = " + vespa_dir + "/lc.csv" + "\n")
    #         f.write("band = J" + "\n")
    #         #TODO cadence
    #         f.write("cadence = " + str(cadence_days) + "\n")
    #         f.write("[constraints]" + "\n")
    #         f.write("maxrad = 12" + "\n")
    #         f.write("secthresh = 1e-4" + "\n")
    #     try:
    #         self.isochrones_starfit(vespa_dir, star_df)
    #         f = FPPCalculation.from_ini(vespa_dir, ini_file="fpp.ini",
    #                                     recalc=True,
    #                                     refit_trap=False,
    #                                     n=20000)
    #         f.trsig.MCMC(refit=True)
    #         f.trsig.save(os.path.join(vespa_dir, 'trsig.pkl'))
    #         trap_corner_file = os.path.join(vespa_dir, 'trap_corner.png')
    #         if not os.path.exists(trap_corner_file) or args.refit_trsig:
    #             f.trsig.corner(outfile=trap_corner_file)
    #         # Including artificial models
    #         boxmodel = BoxyModel(args.artificial_prior, f['pl'].stars.slope.max())
    #         longmodel = LongModel(args.artificial_prior, f['pl'].stars.duration.quantile(0.99))
    #         f.add_population(boxmodel)
    #         f.add_population(longmodel)
    #         f.FPPplots(recalc_lhood=args.recalc_lhood)
    #         logger.info('Re-fitting trapezoid MCMC model...')
    #         f.bootstrap_FPP(1)
    #         for mult, Model in zip(['single', 'binary', 'triple'],
    #                                [StarModel, BinaryStarModel, TripleStarModel]):
    #             starmodel_file = os.path.join(vespa_dir, '{}_starmodel_{}.h5'.format(args.ichrone, mult))
    #             corner_file1 = os.path.join(vespa_dir,
    #                                         '{}_corner_{}_physical.png'.format(args.ichrone, mult))
    #             corner_file2 = os.path.join(vespa_dir,
    #                                         '{}_corner_{}_observed.png'.format(args.ichrone, mult))
    #             if not os.path.exists(corner_file1) or not os.path.exists(corner_file2):
    #                 logger.info('Making StarModel corner plots...')
    #                 starmodel = Model.load_hdf(starmodel_file)
    #                 corner_base = os.path.join(vespa_dir,
    #                                            '{}_corner_{}'.format(args.ichrone, mult))
    #                 starmodel.corner_plots(corner_base)
    #             logger.info('Bootstrap results ({}) written to {}.'.format(1,
    #                                                                        os.path.join(os.path.abspath(vespa_dir),
    #                                                                                     'results_bootstrap.txt')))
    #         logger.info('VESPA FPP calculation successful. ' +
    #                     'Results/plots written to {}.'.format(os.path.abspath(vespa_dir)))
    #         print('VESPA FPP for {}: {}'.format(f.name, f.FPP()))
    #         fpp_df = pandas.DataFrame(columns=['fpp'])
    #         fpp_df.append({'name': f.name, 'fpp': f.FPP()})
    #         fpp_df.to_csv(vespa_dir + "fpp.csv", index=False)
    #     except KeyboardInterrupt:
    #         raise
    #     except:
    #         logger.error('FPP calculation failed for {}.'.format(vespa_dir), exc_info=True)
    #
    # def isochrones_starfit(self, vespa_dir, star_df):
    #     with open(vespa_dir + "star.ini", 'w+') as f:
    #         feh = star_df.at[0, "feh"]
    #         feh_err = star_df.at[0, "feh_err"]
    #         logg = star_df.at[0, "logg"]
    #         logg_err = star_df.at[0, "logg_err"]
    #         j = star_df.at[0, "j"]
    #         j_err = star_df.at[0, "j_err"]
    #         h = star_df.at[0, "h"]
    #         h_err = star_df.at[0, "h_err"]
    #         k = star_df.at[0, "k"]
    #         k_err = star_df.at[0, "k_err"]
    #         kp = star_df.at[0, "kp"]
    #         f.write("Teff = " + str(star_df.at[0, "Teff_star"]) + ", " + str(star_df.at[0, "Teff_star"] * 0.1) + "\n")
    #         if feh is not None and not np.isnan(feh) and feh_err is not None and not np.isnan(feh_err):
    #             f.write("feh = " + str(star_df.at[0, "feh"]) + ", " + str(star_df.at[0, "feh_err"]) + "\n")
    #         elif feh is not None and not np.isnan(feh):
    #             f.write("feh = " + str(star_df.at[0, "feh"]) + "\n")
    #         if logg is not None and not np.isnan(logg) and logg_err is not None and not np.isnan(logg_err):
    #             f.write("logg = " + str(star_df.at[0, "logg"]) + ", " + str(star_df.at[0, "logg_err"]) + "\n")
    #         elif logg is not None and not np.isnan(logg):
    #             f.write("logg = " + str(star_df.at[0, "logg"]) + "\n")
    #         f.write("[sherlock]\n")
    #         if j is not None and not np.isnan(j) and j_err is not None and not np.isnan(j_err):
    #             f.write("J = " + str(star_df.at[0, "j"]) + ", " + str(star_df.at[0, "j_err"]) + "\n")
    #         elif j is not None and not np.isnan(j):
    #             f.write("J = " + str(star_df.at[0, "j"]) + "\n")
    #         if h is not None  and not np.isnan(h) and h_err is not None and not np.isnan(h_err):
    #             f.write("H = " + str(star_df.at[0, "h"]) + ", " + str(star_df.at[0, "h_err"]) + "\n")
    #         elif h is not None and not np.isnan(h):
    #             f.write("H = " + str(star_df.at[0, "h"]) + "\n")
    #         if k is not None and not np.isnan(k) and logg_err is not None and not np.isnan(logg_err):
    #             f.write("K = " + str(star_df.at[0, "k"]) + ", " + str(star_df.at[0, "k_err"]) + "\n")
    #         elif k is not None:
    #             f.write("K = " + str(star_df.at[0, "kp"]) + "\n")
    #         if kp is not None and not np.isnan(kp):
    #             f.write("Kepler = " + str(star_df.at[0, "kp"]) + "\n")
    #     starfit(
    #         vespa_dir,
    #         multiplicities=["single", "binary", "triple"],
    #         models="mist",
    #         use_emcee=True,
    #         plot_only=False,
    #         overwrite=False,
    #         verbose=False,
    #         logger=None,
    #         starmodel_type=None,
    #         skip_initial_state_check=True,
    #         ini_file="star.ini",
    #         no_plots=False,
    #         bands=None
    #     )


class TriceratopsThreadValidator:
    """
    Used to run a single scenario validation with TRICERATOPS
    """
    def __init__(self) -> None:
        super().__init__()

    @staticmethod
    def validate(input):
        """
        Computes the input scenario FPP and NFPP. In addition, FPP2 and FPP3+, from the probability boost proposed in
        Lissauer et al. (2012) eq. 8 and 9 for systems where one or more planets have already been confirmed, are also
        provided just in case they are useful so they don't need to be manually calculated

        :param input: ValidatorInput
        :return: the FPP values, the probabilities dataframe and additional target values
        """
        #input.target.calc_depths(tdepth=input.depth, all_ap_pixels=input.apertures)
        input.target.stars.loc['plx', 0] = 0
        input.target.calc_probs(time=input.time, flux_0=input.flux, flux_err_0=input.sigma, P_orb=float(input.period),
                                contrast_curve_file=input.contrast_curve, parallel=True)
        fpp2 = 1 - 25 * (1 - input.target.FPP) / (25 * (1 - input.target.FPP) + input.target.FPP)
        fpp3 = 1 - 50 * (1 - input.target.FPP) / (50 * (1 - input.target.FPP) + input.target.FPP)
        input.target.probs.to_csv(input.save_dir + "/validation_" + str(input.run) + "_scenarios.csv", index=False)
        input.target.plot_fits(save=True, fname=input.save_dir + "/scenario_" + str(input.run) + "_fits",
                               time=input.time, flux_0=input.flux, flux_err_0=input.sigma)
        return input.target.FPP, input.target.NFPP, fpp2, fpp3, input.target.probs, input.target.star_num, \
               input.target.u1, input.target.u2, input.target.fluxratio_EB, input.target.fluxratio_comp


class ValidatorInput:
    """
    Wrapper class for input arguments of TriceratopsThreadValidator.
    """
    def __init__(self, save_dir, target, time, flux, sigma, period, depth, apertures, run, contrast_curve):
        self.save_dir = save_dir
        self.target = target
        self.time = time
        self.flux = flux
        self.sigma = sigma
        self.period = period
        self.depth = depth
        self.apertures = apertures
        self.run = run
        self.contrast_curve = contrast_curve



class TransitResult:
    """
    Attributes-only class to store the results of transit search from :class:`sherlockpipe.sherlock.Sherlock`
    """
    def __init__(self, power_args, results, period, per_err, duration,
                 t0, t0s, depths, depths_err, depth, depth_err, odd_even_mismatch, depth_mean_even, depth_mean_odd,
                 count, snr, sde, fap, border_score, in_transit, harmonic=None, harmonic_spectrum=None):
        self.power_args = power_args
        self.results = results
        self.period = period
        self.per_err = per_err
        self.duration = duration
        self.t0 = t0
        self.t0s = t0s
        self.depths = depths
        self.depths_err = depths_err
        self.depth = depth
        self.depth_err = depth_err
        self.odd_even_mismatch = odd_even_mismatch
        self.depth_mean_even = depth_mean_even
        self.depth_mean_odd = depth_mean_odd
        self.count = count
        self.snr = snr
        self.sde = sde
        self.fap = fap
        self.border_score = border_score
        self.in_transit = in_transit
        self.harmonic = harmonic
        self.harmonic_spectrum = harmonic_spectrum



import numpy as np

from sherlockpipe.scoring.SignalSelector import SignalSelector, SignalSelection


class BasicSdeSignalSelector(SignalSelector):
    """
    Selects the signal with best SDE
    """
    def __init__(self):
        super().__init__()

    def select(self, id_run, sherlock_target, star_info, transits_min_count, time, lcs, transit_results, wl, cadence):
        detrends_sde = np.nan_to_num([transit_result.sde
                                      for key, transit_result in transit_results.items()])
        best_signal_sde = np.nanmax(detrends_sde)
        best_signal_sde_index = np.nanargmax(detrends_sde)
        selected_signal_snr = transit_results[best_signal_sde_index].snr
        selected_signal = transit_results[best_signal_sde_index]
        if best_signal_sde > sherlock_target.sde_min and selected_signal_snr > sherlock_target.snr_min:  # and SDE[a] > SDE_min and FAP[a] < FAP_max):
            best_signal_score = 1
        else:
            best_signal_score = 0
        return SignalSelection(best_signal_score, best_signal_sde_index, selected_signal)


# import os
import datetime
import os
import pathlib

from PyPDF2 import PdfReader
from astropy.coordinates import Angle
from pdf2image import pdf2image
from reportlab.lib import colors
from reportlab.lib.pagesizes import A4
from reportlab.lib.units import cm
from reportlab.lib.enums import TA_CENTER, TA_JUSTIFY
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
from reportlab.platypus import BaseDocTemplate, PageTemplate, Frame, Paragraph, Spacer, PageBreak, \
    Image, Table, TableStyle, ListFlowable
from os import path
from astropy import units as u
import pandas as pd
import numpy as np

width, height = A4
resources_dir = path.join(path.dirname(__file__))


class ValidationReport:
    """
    This class creates a validation report for the validation module of SHERLOCK.
    """
    LOGO_IMAGE = resources_dir + "/../resources/images/sherlock3.png"

    def __init__(self, data_dir, file_name, object_id, ra, dec, t0, period, duration, depth, v, j, h, k):
        self.data_dir = data_dir
        self.file_name = file_name
        self.object_id = object_id
        self.ra = ra
        self.dec = dec
        self.t0 = t0
        self.period = period
        self.duration = duration
        self.depth = depth
        self.v = v
        self.j = j
        self.h = h
        self.k = k

    @staticmethod
    def row_colors(df, table_object):
        data_len = len(df)
        for each in range(1, data_len + 1):
            if each % 2 == 1:
                bg_color = colors.whitesmoke
            else:
                bg_color = colors.lightgrey
            table_object.setStyle(TableStyle([('BACKGROUND', (0, each), (-1, each), bg_color)]))

    def create_header(self, canvas, doc):
        canvas.saveState()

        # Logo:
        canvas.drawImage(self.LOGO_IMAGE, x=1.5 * cm, y=26.8 * cm, height=2 * cm, width=2 * cm,
                         preserveAspectRatio=True)

        # Title:
        object_id_text = 'SHERLOCK Statistical Validation Report: %s' % self.object_id
        canvas.setFont(psfontname="Helvetica", size=12)
        canvas.drawRightString(x=14.5 * cm, y=27.5 * cm, text=object_id_text)
        if doc.page == 1:
            object_id_text = '%s STATISTICAL VALIDATION REPORT' % self.object_id
            canvas.setFont(psfontname="Helvetica-Bold", size=20)
            canvas.drawCentredString(x=11 * cm, y=25.5 * cm, text=object_id_text)

        # Report date:
        report_date = datetime.datetime.now().strftime("%a, %d %B %Y, %H:%M:%S")
        report_date_text = '%s' % report_date

        canvas.setFont("Helvetica", 9)
        canvas.drawRightString(20.5 * cm, 28 * cm, report_date_text)

        canvas.restoreState()

    def create_footer(self, canvas, doc):
        canvas.saveState()

        # if doc.page == 1:
        #     # Footer con superíndice:
        #     textobject = canvas.beginText()
        #     textobject.setTextOrigin(1.8 * cm, 2.1 * cm)
        #     textobject.setFont("Helvetica", 5)
        #     textobject.setRise(5)
        #     textobject.textOut('1 ')
        #     textobject.setRise(0)
        #     textobject.setFont("Helvetica", 7)
        #     pie_pagina = 'Three possible observability values are defined: 1 - Entire transit is required, ' \
        #                  '0.5 - Transit midtime and either ingress or egress at least are required,\n' \
        #                  '0.25 - Only ingress or egress are required, with moon constraints of % sº as minimum ' \
        #                  'distance for new moon and % sº as minimum distance for full moon\n' \
        #                  'and for the observatories listed in the Table 2.' % (self.min_dist, self.max_dist)
        #
        #     for line in pie_pagina.splitlines():
        #         textobject.textLine(line)
        #
        #     canvas.drawText(textobject)

        # Powered by:
        page = "Powered by TRICERATOPS & ReportLab"
        canvas.setFont("Helvetica", 9)
        canvas.drawRightString(7 * cm, 0.5 * cm, page)

        # Page:
        page = "Page %s" % doc.page
        canvas.setFont("Helvetica", 9)
        canvas.drawRightString(20.5 * cm, 0.5 * cm, page)

        canvas.restoreState()

    def create_report(self):
        # Styles to be used
        styles = getSampleStyleSheet()
        styles.add(ParagraphStyle(name="ParagraphAlignCenter", alignment=TA_CENTER))
        styles.add(ParagraphStyle(name="ParagraphAlignJustify", alignment=TA_JUSTIFY))
        styles.wordWrap = 'LTR'
        table_style = TableStyle([('VALIGN', (0, 0), (-1, -1), 'MIDDLE'),
                                  ('ALIGN', (0, 0), (-1, -1), 'CENTER'),
                                  ('INNERGRID', (0, 0), (-1, -1), 0.25, colors.black),
                                  ('BOX', (0, 0), (-1, -1), 0.25, colors.black),
                                  ('FONTSIZE', (0, 0), (-1, -1), 10),
                                  ])
        table_style_small = TableStyle([('VALIGN', (0, 0), (-1, -1), 'MIDDLE'),
                                  ('ALIGN', (0, 0), (-1, -1), 'CENTER'),
                                  ('INNERGRID', (0, 0), (-1, -1), 0.25, colors.black),
                                  ('BOX', (0, 0), (-1, -1), 0.25, colors.black),
                                  ('FONTSIZE', (0, 0), (-1, -1), 8),
                                  ])
        # Content:
        story = [Spacer(1, 75)]
        introduction = '<font name="HELVETICA" size="9">This document is created by the SHERLOCK statistical validtion report generator (' \
                       '<a href="https://github.com/franpoz/SHERLOCK" color="blue">https://github.com/franpoz/SHERLOCK</a>) ' \
                       'and focuses on the target star %s.</font>' % self.object_id
        story.append(Paragraph(introduction, styles["ParagraphAlignJustify"]))

        story.append(Spacer(1, 30))

        # Generamos la tabla 1 con los parámetros:
        tabla1_data = [['RA (deg)', 'Dec (deg)', 'V (mag)', 'J (mag)', 'H (mag)', 'K (mag)'],
                       [Angle(self.ra, u.deg).to_string(unit=u.hourangle, sep=':',
                                                        precision=2) if self.ra is not None else '-',
                        Angle(self.dec, u.deg).to_string(unit=u.deg, sep=':',
                                                         precision=2) if self.dec is not None else '-',
                        round(self.v, 2) if self.v is not None else '-',
                        round(self.j, 2) if self.j is not None else '-',
                        round(self.h, 2) if self.h is not None else '-',
                        round(self.k, 2) if self.k is not None else '-']]
        table1_colwidth = [3.5 * cm, 3.5 * cm, 2 * cm, 2 * cm, 2 * cm, 2 * cm]
        table1_number_rows = len(tabla1_data)
        tabla1 = Table(tabla1_data, table1_colwidth, table1_number_rows * [0.75 * cm])
        tabla1.setStyle(table_style)
        # Le damos el estilo alternando colores de filas:
        ValidationReport.row_colors(tabla1_data, tabla1)
        story.append(tabla1)
        table1_descripcion = '<font name="HELVETICA" size="9"><strong>Table 1: </strong>\
                        The proposed target parameters.</font>'
        story.append(Spacer(1, 5))
        story.append(Paragraph(table1_descripcion, styles["ParagraphAlignCenter"]))
        story.append(Spacer(1, 15))
        # Generamos la tabla 2 con los parámetros:
        tabla2_data = [['T0 (d)', 'Period (d)', 'Duration (h)', 'Depth (ppt)'],
                       [round(self.t0, 4),
                        round(self.period, 4),
                        round(self.duration / 60, 2),
                        round(self.depth, 3)]]
        table2_colwidth = [4 * cm, 4 * cm, 3.5 * cm, 3.5 * cm]
        table2_number_rows = len(tabla2_data)
        tabla2 = Table(tabla2_data, table2_colwidth, table2_number_rows * [0.75 * cm])
        tabla2.setStyle(table_style)
        ValidationReport.row_colors(tabla2_data, tabla2)
        story.append(tabla2)
        story.append(Spacer(1, 5))
        table2_descripcion = '<font name="HELVETICA" size="9"><strong>Table 2: </strong>' \
                             'The candidate parameters.</font>'
        story.append(Paragraph(table2_descripcion, styles["ParagraphAlignCenter"]))
        story.append(Spacer(1, 15))
        folded_curve_file = self.data_dir + "/folded_curve.png"
        figure = 1
        if os.path.exists(folded_curve_file):
            story.append(Image(folded_curve_file, width=16 * cm, height=10 * cm))
            descripcion = '<font name="HELVETICA" size="9"><strong>Figure ' + str(
                figure) + ': </strong>Folded curve of the found candidate.</font>'
            story.append(Spacer(1, 5))
            story.append(Paragraph(descripcion, styles["ParagraphAlignCenter"]))
            story.append(Spacer(1, 15))
            figure = figure + 1
        contrast_curve_file = self.data_dir + "/contrast_curve.png"
        if os.path.exists(contrast_curve_file):
            story.append(Image(contrast_curve_file, width=16 * cm, height=10 * cm))
            descripcion = '<font name="HELVETICA" size="9"><strong>Figure ' + str(
                figure) + ': </strong>Contrast curve for target.</font>'
            story.append(Spacer(1, 5))
            story.append(Paragraph(descripcion, styles["ParagraphAlignCenter"]))
            story.append(Spacer(1, 15))
            figure = figure + 1
        all_files = os.listdir(self.data_dir)
        fov_file = self.data_dir + '/fov.png'
        for file in all_files:
            if file.startswith('field_'):
                images = pdf2image.convert_from_path(self.data_dir + '/' + file)
                for i in range(len(images)):
                    # Save pages as images in the pdf
                    images[i].save(fov_file, 'PNG')
                break
        story.append(Image(fov_file, width=16 * cm, height=7 * cm))
        descripcion = '<font name="HELVETICA" size="9"><strong>Figure ' + str(
            figure) + ': </strong>Nearby stars for target and its aperture</font>'
        story.append(Spacer(1, 5))
        story.append(Paragraph(descripcion, styles["ParagraphAlignCenter"]))
        story.append(Spacer(1, 15))
        figure = figure + 1
        validation_file = self.data_dir + "/validation_scenarios.csv"
        if os.path.exists(validation_file):
            table_data = [['ID', 'scenario', 'M_s', 'R_s', 'P_orb', 'inc', 'b', 'ecc', 'w', 'R_p', 'M_EB', 'R_EB',
                           'prob']]
            metrics_df = pd.read_csv(validation_file)
            for index, metric_row in metrics_df.iterrows():
                table_data.append([str(metric_row['ID']),
                                   metric_row['scenario'],
                                   round(metric_row['M_s'], 2),
                                   round(metric_row['R_s'], 2),
                                   round(metric_row['P_orb'], 2),
                                   round(metric_row['inc'], 2),
                                   round(metric_row['b'], 2),
                                   round(metric_row['ecc'], 2),
                                   round(metric_row['w'], 2),
                                   round(metric_row['R_p'], 2),
                                   round(metric_row['M_EB'], 2),
                                   round(metric_row['R_EB'], 2),
                                   round(metric_row['prob'], 6)])
            table_colwidth = [2.3 * cm, 2 * cm, 1 * cm, 1 * cm, 1 * cm, 1 * cm, 1 * cm, 1 * cm, 1 * cm, 1 * cm,
                              1 * cm, 1 * cm, 2.5 * cm]
            table_number_rows = len(table_data)
            table = Table(table_data, table_colwidth, table_number_rows * [0.5 * cm])
            table.setStyle(table_style_small)
            ValidationReport.row_colors(metrics_df, table)
            story.append(table)
            story.append(Spacer(1, 5))
            table_descripcion = '<font name="HELVETICA" size="9"><strong>Table 3: </strong>' \
                                'Scenarios attributes and probabilities.</font>'
            story.append(Paragraph(table_descripcion, styles["ParagraphAlignCenter"]))
            story.append(Spacer(1, 15))
        validation_file = self.data_dir + "/validation.csv"
        if os.path.exists(validation_file):
            table_data = [['Scenario', 'FPP', 'NFPP', 'FPP2', 'FPP3+']]
            metrics_df = pd.read_csv(validation_file)
            for index, metric_row in metrics_df.iterrows():
                table_data.append([metric_row['scenario'],
                                   round(metric_row['FPP'], 6),
                                   round(metric_row['NFPP'], 6),
                                   round(metric_row['FPP2'], 6),
                                   round(metric_row['FPP3+'], 6)])
            table_colwidth = [4 * cm, 4 * cm, 3.5 * cm]
            table_number_rows = len(table_data)
            table = Table(table_data, table_colwidth, table_number_rows * [0.5 * cm])
            table.setStyle(table_style)
            ValidationReport.row_colors(metrics_df, table)
            story.append(table)
            story.append(Spacer(1, 5))
            table_descripcion = '<font name="HELVETICA" size="9"><strong>Table 4: </strong>' \
                                'Validation results.</font>'
            story.append(Paragraph(table_descripcion, styles["ParagraphAlignCenter"]))
            story.append(Spacer(1, 15))
        result_map_file = self.data_dir + "/triceratops_map.png"
        if os.path.exists(result_map_file):
            story.append(Image(result_map_file, width=9 * cm, height=9 * cm))
            descripcion = '<font name="HELVETICA" size="9"><strong>Figure ' + str(
                figure) + ': </strong>Validation map for the mean scenario.</font>'
            story.append(Spacer(1, 5))
            story.append(Paragraph(descripcion, styles["ParagraphAlignCenter"]))
            story.append(Spacer(1, 15))
            figure = figure + 1
        scenario = 0
        scenarios_file = self.data_dir + "/scenario_" + str(scenario) + "_fits.pdf"
        scenarios_png_file = self.data_dir + "/scenario_" + str(scenario) + "_fits.png"
        while os.path.exists(scenarios_file):
            images = pdf2image.convert_from_path(scenarios_file)
            for i in range(len(images)):
                # Save pages as images in the pdf
                images[i].save(scenarios_png_file, 'PNG')
            story.append(Image(scenarios_png_file, width=16 * cm, height=20 * cm))
            descripcion = '<font name="HELVETICA" size="9"><strong>Figure ' + str(
                figure) + ': </strong>Best fits for scenario no. .' + str(scenario) + '</font>'
            story.append(Spacer(1, 5))
            story.append(Paragraph(descripcion, styles["ParagraphAlignCenter"]))
            story.append(Spacer(1, 15))
            figure = figure + 1
            scenario = scenario + 1
            scenarios_file = self.data_dir + "/scenario_" + str(scenario) + "_fits.pdf"
            scenarios_png_file = self.data_dir + "/scenario_" + str(scenario) + "_fits.png"
        # neighbours_file_index = 0
        # neighbours_file = self.data_dir + "/star_nb_" + (neighbours_file_index) + ".png"
        # figure = 1
        # while os.path.exists(neighbours_file):
        #     story.append(Image(neighbours_file, width=16 * cm, height=16 * cm))
        #     descripcion = '<font name="HELVETICA" size="9"><strong>Figure ' + str(figure) + ': </strong>' \
        #                          'Nearby stars folded plot with the same period than the candidate.</font>'
        #     story.append(Spacer(1, 5))
        #     story.append(Paragraph(descripcion, styles["ParagraphAlignCenter"]))
        #     story.append(Spacer(1, 15))
        #     neighbours_file_index = neighbours_file_index + 1
        #     neighbours_file = self.data_dir + "/star_nb_" + str(neighbours_file_index) + ".png"
        #     figure = figure + 1

        # Construimos el documento:
        global_frame = Frame(1.5 * cm, 1.1 * cm, 18 * cm, 25.4 * cm, id='normal', showBoundary=0)
        global_template = PageTemplate(id='UnaColumna', frames=global_frame,
                                       onPage=self.create_header, onPageEnd=self.create_footer)
        doc = BaseDocTemplate(self.data_dir + "/" + self.object_id + "_" + self.file_name, pagesize=A4,
                              rightMargin=40, leftMargin=40,
                              topMargin=95, bottomMargin=15,
                              pageTemplates=global_template)
        doc.build(story)



from argparse import ArgumentParser

from sherlockpipe.vetting.run import run_vet

if __name__ == '__main__':
    ap = ArgumentParser(description='Vetting of Sherlock objects of interest')
    ap.add_argument('--object_dir', help="If the object directory is not your current one you need to provide the "
                                         "ABSOLUTE path", required=False)
    ap.add_argument('--candidate', type=int, default=None, help="The candidate signal to be used.", required=False)
    ap.add_argument('--ml', action='store_true', default=False, help="Whether tu run WATSON-NET.", required=False)
    ap.add_argument('--properties', help="The YAML file to be used as input.", required=False)
    ap.add_argument('--cpus', type=int, default=None, help="The number of CPU cores to be used.", required=False)
    args = ap.parse_args()
    run_vet(args.object_dir, args.candidate, args.properties, args.cpus, run_iatson=args.ml)



from typing import Dict, List

from spock import FeatureClassifier, DeepRegressor
from sherlockpipe.system_stability.stability_calculator import StabilityCalculator, SimulationInput
import pandas as pd


class SpockStabilityCalculator(StabilityCalculator):
    """
    Runs the stability computation by computing the stability probability and the median expected instability time for
    each scenario
    """
    def run_simulation(self, simulation_input: SimulationInput) -> dict:
        """
        Runs one stability scenario using Spock.

        :param SimulationInput simulation_input:
        :return dict: the result with the spock-specific metrics stability_probability and the median_expected_instability_time
        """
        sim = self.init_rebound_simulation(simulation_input)
        feature_classifier_model = FeatureClassifier()
        deep_regressor_model = DeepRegressor()
        stability_probability = feature_classifier_model.predict_stable(sim)
        median, lower, upper = deep_regressor_model.predict_instability_time(sim, samples=10000)
        return {"star_mass": simulation_input.star_mass,
                "periods": ",".join([str(planet_period) for planet_period in simulation_input.planet_periods]),
                "masses": ",".join([str(mass_value) for mass_value in simulation_input.mass_arr]),
                "inclinations": ",".join([str(ecc_value) for ecc_value in simulation_input.inc_arr]),
                "eccentricities": ",".join([str(ecc_value) for ecc_value in simulation_input.ecc_arr]),
                "arg_periastron": ",".join([str(ecc_value) for ecc_value in simulation_input.omega_arr]),
                "stability_probability": stability_probability, "median_expected_instability_time": median}

    def store_simulation_results(self, simulation_results: List[Dict], results_dir: str):
        """
        Stores the spock results in the given directory

        :param List[Dict] simulation_results: the list of results dictionaries
        :param str results_dir: directory to store the file
        """
        result_file = results_dir + "/stability_spock.csv"
        results_df = pd.DataFrame(columns=['star_mass', 'periods', 'masses', 'inclinations', 'eccentricities',
                                           'arg_periastron', 'stability_probability',
                                           'median_expected_instability_time'])
        results_df = results_df.append(simulation_results, ignore_index=True)
        results_df = results_df.sort_values('stability_probability', ascending=False)
        results_df.to_csv(result_file, index=False)



import logging
import os
import shutil
import sys
from pathlib import Path

import pandas as pd

from sherlockpipe.loading import common
from sherlockpipe.vetting.vetter import Vetter


def run_vet(object_dir, candidate, properties, cpus=os.cpu_count() - 1, run_iatson=False):
    object_dir = os.getcwd() if object_dir is None else object_dir
    candidates = pd.read_csv(object_dir + "/candidates.csv")
    if candidate is not None:
        vetting_dir = object_dir + "/vet_" + str(candidate)
    else:
        vetting_dir = object_dir + "/vet_" + str(Path(properties).stem)
    if os.path.exists(vetting_dir) or os.path.isdir(vetting_dir):
        shutil.rmtree(vetting_dir, ignore_errors=True)
    os.mkdir(vetting_dir)
    vetter = Vetter(object_dir, vetting_dir, candidate is not None, candidates)
    file_dir = vetter.watson.object_dir + "/vetting.log"
    if os.path.exists(file_dir):
        os.remove(file_dir)
    if not isinstance(logging.root, logging.RootLogger):
        logging.root = logging.RootLogger(logging.INFO)
    formatter = logging.Formatter(fmt='%(asctime)s %(levelname)-8s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    logger = logging.getLogger()
    while len(logger.handlers) > 0:
        logger.handlers.pop()
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    handler = logging.FileHandler(file_dir)
    handler.setLevel(logging.INFO)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logging.info("Starting vetting")
    star_df = pd.read_csv(vetter.object_dir() + "/params_star.csv")
    transits_df = None
    if candidate is None:
        user_properties = common.load_from_yaml(properties)
        candidate = pd.DataFrame(columns=['id', 'period', 'depth', 't0', 'sectors', 'number', 'lc'])
        candidate = candidate.append(user_properties, ignore_index=True)
        candidate['id'] = star_df.iloc[0]["obj_id"]
    else:
        candidate_selection = int(candidate)
        if candidate_selection < 1 or candidate_selection > len(candidates.index):
            raise SystemExit("User selected a wrong candidate number.")
        candidates = candidates.rename(columns={'Object Id': 'id'})
        candidates['number'] = 1
        candidate = candidates.iloc[[candidate_selection - 1]]
        candidate.iloc[:]['number'] = candidate_selection
        transits_df_file = vetter.object_dir() + "/transits_stats.csv"
        if os.path.exists(transits_df_file):
            transits_df = pd.read_csv(vetter.object_dir() + "/transits_stats.csv")
            transits_df = transits_df[transits_df["candidate"] == candidate_selection - 1]
            if len(transits_df) == 0:
                logging.info("Not NAN transits found for candidate in transits_stats.csv file")
                transits_df = None
        # watson.data_dir = watson.object_dir
        logging.info("Selected signal number " + str(candidate_selection))
    transits_mask = []
    for i in range(0, int(candidate['number']) - 1):
        transits_mask.append({"P": candidates.iloc[i]["period"], "T0": candidates.iloc[i]["t0"],
                              "D": candidates.iloc[i]["duration"] * 2})
    vetter.run(cpus, candidate=candidate, star_df=star_df.iloc[0], transits_df=transits_df, transits_mask=transits_mask,
               iatson_enabled=run_iatson)
